# ActivitySim
# See full license in LICENSE.txt.
import logging
from builtins import range

import numpy as np
import pandas as pd

from activitysim.abm.models.util import estimation
from activitysim.abm.models.util.trip import cleanup_failed_trips, failed_trip_cohorts
from activitysim.core import chunk, config, inject, logit, pipeline, tracing
from activitysim.core.util import reindex

from .util import probabilistic_scheduling as ps

logger = logging.getLogger(__name__)

"""
StopDepartArrivePeriodModel

StopDepartArriveProportions.csv
tourpurp,isInbound,interval,trip,p1,p2,p3,p4,p5...p40

"""

NO_TRIP_ID = 0
NO_DEPART = 0

DEPART_ALT_BASE = "DEPART_ALT_BASE"

FAILFIX = "FAILFIX"
FAILFIX_CHOOSE_MOST_INITIAL = "choose_most_initial"
FAILFIX_DROP_AND_CLEANUP = "drop_and_cleanup"
FAILFIX_DEFAULT = FAILFIX_CHOOSE_MOST_INITIAL

DEPARTURE_MODE = "departure"
DURATION_MODE = "stop_duration"
PROBS_JOIN_COLUMNS_DEPARTURE_BASED = [
    "primary_purpose",
    "outbound",
    "tour_hour",
    "trip_num",
]
PROBS_JOIN_COLUMNS_DURATION_BASED = ["outbound", "stop_num"]


def set_tour_hour(trips, tours):
    """
    add columns 'tour_hour', 'earliest', 'latest' to trips

    Parameters
    ----------
    trips: pd.DataFrame
    tours: pd.DataFrame

    Returns
    -------
    modifies trips in place
    """

    # all trips must depart between tour start and end
    trips["earliest"] = reindex(tours.start, trips.tour_id)
    trips["latest"] = reindex(tours.end, trips.tour_id)

    # tour_hour is start for outbound trips, and end for inbound trips
    trips["tour_hour"] = np.where(
        trips.outbound, trips["earliest"], trips["latest"]
    ).astype(np.int8)

    # subtours indexed by parent_tour_id
    subtours = tours.loc[
        tours.primary_purpose == "atwork",
        ["tour_num", "tour_count", "parent_tour_id", "start", "end"],
    ]

    subtours.parent_tour_id = subtours.parent_tour_id.astype(np.int64)
    subtours = subtours.set_index("parent_tour_id")
    subtours = subtours.astype(np.int16)  # remaining columns are all small ints

    # bool series
    trip_has_subtours = trips.tour_id.isin(subtours.index)

    outbound = trip_has_subtours & trips.outbound
    trips.loc[outbound, "latest"] = reindex(
        subtours[subtours.tour_num == 1]["start"], trips[outbound].tour_id
    )

    inbound = trip_has_subtours & ~trips.outbound
    trips.loc[inbound, "earliest"] = reindex(
        subtours[subtours.tour_num == subtours.tour_count]["end"],
        trips[inbound].tour_id,
    )


def set_stop_num(trips):
    """
    Convert trip_num to stop_num in order to work with duration-based
    probs that are keyed on stop num. For outbound trips, trip n chooses
    the duration of stop n-1 (the trip origin). For inbound trips, trip n
    chooses the duration of stop n (the trip destination). This means
    outbound trips technically choose a departure time while inbound trips
    choose an arrival.
    """
    trips["stop_num"] = trips["trip_num"] - 1
    trips["stop_num"] = trips.stop_num.where(trips["outbound"], trips["trip_num"])


def update_tour_earliest(trips, outbound_choices):
    """
    Updates "earliest" column for inbound trips based on
    the maximum outbound trip departure time of the tour.
    This is done to ensure inbound trips do not depart
    before the last outbound trip of a tour.

    Parameters
    ----------
    trips: pd.DataFrame
    outbound_choices: pd.Series
        time periods depart choices, one per trip (except for trips with
        zero probs)
    Returns
    -------
    modifies trips in place
    """
    # append outbound departure times to trips
    tmp_trips = trips.copy()
    tmp_trips["outbound_departure"] = outbound_choices.reindex(tmp_trips.index)

    # get max outbound trip departure times for all person-tours
    max_outbound_person_departures = tmp_trips.groupby(["person_id", "tour_id"])[
        "outbound_departure"
    ].max()
    max_outbound_person_departures.name = "max_outbound_departure"

    # append max outbound trip departure times to trips
    tmp_trips = tmp_trips.merge(
        max_outbound_person_departures,
        left_on=["person_id", "tour_id"],
        right_index=True,
    )

    # set the trips "earliest" column equal to the max outbound departure
    # time for all inbound trips. preserve values that were used for outbound trips
    tmp_trips["earliest"] = tmp_trips["earliest"].where(
        tmp_trips["outbound"], tmp_trips["max_outbound_departure"]
    )

    trips["earliest"] = tmp_trips["earliest"].reindex(trips.index)

    return


def schedule_trips_in_leg(
    outbound,
    trips,
    probs_spec,
    model_settings,
    is_last_iteration,
    trace_hh_id,
    trace_label,
):
    """

    Parameters
    ----------
    outbound
    trips
    probs_spec
    depart_alt_base
    is_last_iteration
    trace_hh_id
    trace_label

    Returns
    -------
    choices: pd.Series
        depart choice for trips, indexed by trip_id
    """

    failfix = model_settings.get(FAILFIX, FAILFIX_DEFAULT)
    depart_alt_base = model_settings.get("DEPART_ALT_BASE", 0)
    scheduling_mode = model_settings.get("scheduling_mode", "departure")

    if scheduling_mode == "departure":
        probs_join_cols = model_settings.get(
            "probs_join_cols", PROBS_JOIN_COLUMNS_DEPARTURE_BASED
        )
    elif scheduling_mode == "stop_duration":
        probs_join_cols = model_settings.get(
            "probs_join_cols", PROBS_JOIN_COLUMNS_DURATION_BASED
        )
    else:
        logger.error(
            "Invalid scheduling mode specified: {0}.".format(scheduling_mode),
            "Please select one of ['departure', 'stop_duration'] and try again.",
        )

    # logger.debug("%s scheduling %s trips" % (trace_label, trips.shape[0]))

    assert len(trips) > 0
    assert (trips.outbound == outbound).all()

    result_list = []

    # trips to/from tour origin or atwork get tour_hour departure times
    # no need to schedule them if there are no intermediate stops
    to_from_tour_orig = (
        (trips.trip_num == 1) if outbound else (trips.trip_num == trips.trip_count)
    )
    do_not_schedule = to_from_tour_orig | (trips.primary_purpose == "atwork")
    choices = trips.tour_hour[do_not_schedule]

    if do_not_schedule.all():
        return choices

    result_list.append(choices)
    trips = trips[~do_not_schedule]

    # add next_trip_id temp column, and specificy departure constraint column to update
    trips = trips.sort_index()
    if outbound or scheduling_mode == DURATION_MODE:
        trips["next_trip_id"] = np.roll(trips.index, -1)
        is_final = trips.trip_num == trips.trip_count
        # each trip's depart constrains next trip's earliest depart option
        ADJUST_NEXT_DEPART_COL = "earliest"
    else:
        trips["next_trip_id"] = np.roll(trips.index, 1)
        is_final = trips.trip_num == 1
        # if inbound, we are handling in reverse order, so each choice
        # constrains latest depart of the preceding trip
        ADJUST_NEXT_DEPART_COL = "latest"
    trips.next_trip_id = trips.next_trip_id.where(~is_final, NO_TRIP_ID)

    first_trip_in_leg = True
    for i in range(trips.trip_num.min(), trips.trip_num.max() + 1):

        if outbound or scheduling_mode == DURATION_MODE:
            # iterate in ascending trip_num order
            nth_trips = trips[trips.trip_num == i]
        else:
            # iterate over inbound trips in descending trip_num order, skipping the final trip
            nth_trips = trips[trips.trip_num == trips.trip_count - i]

        nth_trace_label = tracing.extend_trace_label(trace_label, "num_%s" % i)

        choices = ps.make_scheduling_choices(
            nth_trips,
            scheduling_mode,
            probs_spec,
            probs_join_cols,
            depart_alt_base,
            first_trip_in_leg=first_trip_in_leg,
            report_failed_trips=is_last_iteration,
            trace_hh_id=trace_hh_id,
            trace_label=nth_trace_label,
        )

        # most initial departure (when no choice was made because all probs were zero)
        if is_last_iteration and (failfix == FAILFIX_CHOOSE_MOST_INITIAL):
            choices = choices.reindex(nth_trips.index)
            logger.warning(
                "%s coercing %s depart choices to most initial"
                % (nth_trace_label, choices.isna().sum())
            )
            choices = choices.fillna(trips[ADJUST_NEXT_DEPART_COL])

        # adjust allowed depart range of next trip
        has_next_trip = nth_trips.next_trip_id != NO_TRIP_ID
        if has_next_trip.any():
            next_trip_ids = nth_trips.next_trip_id[has_next_trip]
            # patch choice any trips with next_trips that weren't scheduled
            trips.loc[next_trip_ids, ADJUST_NEXT_DEPART_COL] = (
                choices.reindex(next_trip_ids.index)
                .fillna(trips[ADJUST_NEXT_DEPART_COL])
                .values
            )

        result_list.append(choices)

        chunk.log_df(trace_label, f"result_list", result_list)

        first_trip_in_leg = False

    if len(result_list) > 1:
        choices = pd.concat(result_list)

    return choices


def run_trip_scheduling(
    trips_chunk,
    tours,
    probs_spec,
    model_settings,
    estimator,
    is_last_iteration,
    chunk_size,
    trace_hh_id,
    trace_label,
):

    set_tour_hour(trips_chunk, tours)
    set_stop_num(trips_chunk)

    # only non-initial trips require scheduling, segment handing first such trip in tour will use most space
    # is_outbound_chooser = (trips.trip_num > 1) & trips.outbound & (trips.primary_purpose != 'atwork')
    # is_inbound_chooser = (trips.trip_num < trips.trip_count) & ~trips.outbound & (trips.primary_purpose != 'atwork')
    # num_choosers = (is_inbound_chooser | is_outbound_chooser).sum()

    result_list = []

    if trips_chunk.outbound.any():
        leg_chunk = trips_chunk[trips_chunk.outbound]
        leg_trace_label = tracing.extend_trace_label(trace_label, "outbound")
        choices = schedule_trips_in_leg(
            outbound=True,
            trips=leg_chunk,
            probs_spec=probs_spec,
            model_settings=model_settings,
            is_last_iteration=is_last_iteration,
            trace_hh_id=trace_hh_id,
            trace_label=leg_trace_label,
        )
        result_list.append(choices)

        chunk.log_df(trace_label, f"result_list", result_list)

        # departure time of last outbound trips must constrain
        # departure times for initial inbound trips
        update_tour_earliest(trips_chunk, choices)

    if (~trips_chunk.outbound).any():
        leg_chunk = trips_chunk[~trips_chunk.outbound]
        leg_trace_label = tracing.extend_trace_label(trace_label, "inbound")
        choices = schedule_trips_in_leg(
            outbound=False,
            trips=leg_chunk,
            probs_spec=probs_spec,
            model_settings=model_settings,
            is_last_iteration=is_last_iteration,
            trace_hh_id=trace_hh_id,
            trace_label=leg_trace_label,
        )
        result_list.append(choices)

        chunk.log_df(trace_label, f"result_list", result_list)

    choices = pd.concat(result_list)

    return choices


@inject.step()
def trip_scheduling(trips, tours, chunk_size, trace_hh_id):

    """
    Trip scheduling assigns depart times for trips within the start, end limits of the tour.

    The algorithm is simplistic:

    The first outbound trip starts at the tour start time, and subsequent outbound trips are
    processed in trip_num order, to ensure that subsequent trips do not depart before the
    trip that preceeds them.

    Inbound trips are handled similarly, except in reverse order, starting with the last trip,
    and working backwards to ensure that inbound trips do not depart after the trip that
    succeeds them.

    The probability spec assigns probabilities for depart times, but those possible departs must
    be clipped to disallow depart times outside the tour limits, the departs of prior trips, and
    in the case of work tours, the start/end times of any atwork subtours.

    Scheduling can fail if the probability table assigns zero probabilities to all the available
    depart times in a trip's depart window. (This could be avoided by giving every window a small
    probability, rather than zero, but the existing mtctm1 prob spec does not do this. I believe
    this is due to the its having been generated from a small household travel survey sample
    that lacked any departs for some time periods.)

    Rescheduling the trips that fail (along with their inbound or outbound leg-mates) can sometimes
    fix this problem, if it was caused by an earlier trip's depart choice blocking a subsequent
    trip's ability to schedule a depart within the resulting window. But it can also happen if
    a tour is very short (e.g. one time period) and the prob spec having a zero probability for
    that tour hour.

    Therefore we need to handle trips that could not be scheduled. There are two ways (at least)
    to solve this problem:

    1) choose_most_initial
    simply assign a depart time to the trip, even if it has a zero probability. It makes
    most sense, in this case, to assign the 'most initial' depart time, so that subsequent trips
    are minimally impacted. This can be done in the final iteration, thus affecting only the
    trips that could no be scheduled by the standard approach

    2) drop_and_cleanup
    drop trips that could no be scheduled, and adjust their leg mates, as is done for failed
    trips in trip_destination.

    Which option is applied is determined by the FAILFIX model setting

    """
    trace_label = "trip_scheduling"
    model_settings_file_name = "trip_scheduling.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    trips_df = trips.to_frame()
    tours = tours.to_frame()

    # add columns 'tour_hour', 'earliest', 'latest' to trips
    set_tour_hour(trips_df, tours)

    # trip_scheduling is a probabilistic model ane we don't support estimation,
    # but we do need to override choices in estimation mode
    estimator = estimation.manager.begin_estimation("trip_scheduling")
    if estimator:
        estimator.write_spec(model_settings, tag="PROBS_SPEC")
        estimator.write_model_settings(model_settings, model_settings_file_name)
        chooser_cols_for_estimation = [
            "person_id",
            "household_id",
            "tour_id",
            "trip_num",
            "trip_count",
            "primary_purpose",
            "outbound",
            "earliest",
            "latest",
            "tour_hour",
        ]
        estimator.write_choosers(trips_df[chooser_cols_for_estimation])

    probs_spec = pd.read_csv(
        config.config_file_path("trip_scheduling_probs.csv"), comment="#"
    )
    # FIXME for now, not really doing estimation for probabilistic model - just overwriting choices
    # besides, it isn't clear that named coefficients would be helpful if we had some form of estimation
    # coefficients_df = simulate.read_model_coefficients(model_settings)
    # probs_spec = map_coefficients(probs_spec, coefficients_df)

    # add tour-based chunk_id so we can chunk all trips in tour together
    trips_df["chunk_id"] = reindex(
        pd.Series(list(range(len(tours))), tours.index), trips_df.tour_id
    )

    assert "DEPART_ALT_BASE" in model_settings
    failfix = model_settings.get(FAILFIX, FAILFIX_DEFAULT)

    max_iterations = model_settings.get("MAX_ITERATIONS", 1)
    assert max_iterations > 0

    choices_list = []

    for (
        chunk_i,
        trips_chunk,
        chunk_trace_label,
    ) in chunk.adaptive_chunked_choosers_by_chunk_id(
        trips_df, chunk_size, trace_label, trace_label
    ):

        i = 0
        while (i < max_iterations) and not trips_chunk.empty:

            # only chunk log first iteration since memory use declines with each iteration
            with chunk.chunk_log(trace_label) if i == 0 else chunk.chunk_log_skip():

                i += 1
                is_last_iteration = i == max_iterations

                trace_label_i = tracing.extend_trace_label(trace_label, "i%s" % i)
                logger.info(
                    "%s scheduling %s trips within chunk %s",
                    trace_label_i,
                    trips_chunk.shape[0],
                    chunk_i,
                )

                choices = run_trip_scheduling(
                    trips_chunk,
                    tours,
                    probs_spec,
                    model_settings,
                    estimator=estimator,
                    is_last_iteration=is_last_iteration,
                    chunk_size=chunk_size,
                    trace_hh_id=trace_hh_id,
                    trace_label=trace_label_i,
                )

                # boolean series of trips whose individual trip scheduling failed
                failed = choices.reindex(trips_chunk.index).isnull()
                logger.info("%s %s failed", trace_label_i, failed.sum())

                if not is_last_iteration:
                    # boolean series of trips whose leg scheduling failed
                    failed_cohorts = failed_trip_cohorts(trips_chunk, failed)
                    trips_chunk = trips_chunk[failed_cohorts]
                    choices = choices[~failed_cohorts]

                choices_list.append(choices)

    trips_df = trips.to_frame()

    choices = pd.concat(choices_list)
    choices = choices.reindex(trips_df.index)

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(
            choices, "trips", "depart"
        )  # override choices
        estimator.write_override_choices(choices)
        estimator.end_estimation()
        assert not choices.isnull().any()

    if choices.isnull().any():
        logger.warning(
            "%s of %s trips could not be scheduled after %s iterations"
            % (choices.isnull().sum(), trips_df.shape[0], i)
        )

        if failfix != FAILFIX_DROP_AND_CLEANUP:
            raise RuntimeError(
                "%s setting '%s' not enabled in settings"
                % (FAILFIX, FAILFIX_DROP_AND_CLEANUP)
            )

        trips_df["failed"] = choices.isnull()
        trips_df = cleanup_failed_trips(trips_df)
        choices = choices.reindex(trips_df.index)

    trips_df["depart"] = choices

    assert not trips_df.depart.isnull().any()

    pipeline.replace_table("trips", trips_df)



import importlib
import itertools
import logging
import os
import re
from pathlib import Path

import larch
import numpy as np
import pandas as pd
import yaml
from larch import DataFrames, Model, P, X
from larch.log import logger_name
from larch.model.model_group import ModelGroup
from larch.util import Dict

from ...abm.models.util import cdap
from .general import apply_coefficients, explicit_value_parameters

_logger = logging.getLogger(logger_name)


def generate_alternatives(n_persons):
    """
    Generate a dictionary of CDAP alternatives.

    The keys are the names of the patterns, and
    the values are the alternative code numbers.

    Parameters
    ----------
    n_persons : int

    Returns
    -------
    Dict
    """
    basic_patterns = ["M", "N", "H"]
    alt_names = list(
        "".join(i) for i in itertools.product(basic_patterns, repeat=n_persons)
    )
    alt_codes = np.arange(1, len(alt_names) + 1)
    return Dict(zip(alt_names, alt_codes))


def apply_replacements(expression, prefix, tokens):
    """
    Convert general person terms to specific person terms for the CDAP model.

    Parameters
    ----------
    expression : str
        An expression from the "Expression" column
        of cdap_INDIV_AND_HHSIZE1_SPEC.csv, or similar.
    prefix : str
        A prefix to attach to each token in `expression`.
    tokens : list-like of str
        A list of tokens to edit within an expression.

    Returns
    -------
    expression : str
        The modified expression
    """
    for i in tokens:
        expression = re.sub(rf"\b{i}\b", f"{prefix}_{i}", expression)
    return expression


def cdap_base_utility_by_person(model, n_persons, spec, alts=None, value_tokens=()):
    """
    Build the base utility by person for each pattern.

    Parameters
    ----------
    model : larch.Model
    n_persons : int
    spec : pandas.DataFrame
        The base utility by person spec provided by
        the ActivitySim framework.
    alts : dict, optional
        The keys are the names of the patterns, and
        the values are the alternative code numbers,
        as created by `generate_alternatives`.  If not
        given, the alts are automatically regenerated
        using that function.
    value_tokens : list-like of str, optional
        A list of tokens to edit within an the expressions,
        generally the column names of the provided values
        from the estimation data bundle.  Only used when
        `n_persons` is more than 1.
    """
    if n_persons == 1:
        for i in spec.index:
            if not pd.isna(spec.loc[i, "M"]):
                model.utility_co[1] += X(spec.Expression[i]) * P(spec.loc[i, "M"])
            if not pd.isna(spec.loc[i, "N"]):
                model.utility_co[2] += X(spec.Expression[i]) * P(spec.loc[i, "N"])
            if not pd.isna(spec.loc[i, "H"]):
                model.utility_co[3] += X(spec.Expression[i]) * P(spec.loc[i, "H"])
    else:
        if alts is None:
            alts = generate_alternatives(n_persons)
        person_numbers = range(1, n_persons + 1)
        for pnum in person_numbers:
            for i in spec.index:
                for aname, anum in alts.items():
                    z = pnum - 1
                    if not pd.isna(spec.loc[i, aname[z]]):
                        x = apply_replacements(
                            spec.Expression[i], f"p{pnum}", value_tokens
                        )
                        model.utility_co[anum] += X(x) * P(spec.loc[i, aname[z]])


def interact_pattern(n_persons, select_persons, tag):
    """
    Compile a regex pattern to match CDAP alternatives.

    Parameters
    ----------
    n_persons : int
    select_persons : list-like of int
        The persons to be selected.
    tag : str
        The activity letter, currently one of {M,N,H}.

    Returns
    -------
    re.compile
    """

    pattern = ""
    p = 1
    while len(pattern) < n_persons:
        pattern += tag if p in select_persons else "."
        p += 1
    return re.compile(pattern)


def cdap_interaction_utility(model, n_persons, alts, interaction_coef, coefficients):

    person_numbers = list(range(1, n_persons + 1))

    matcher = re.compile("coef_[HMN]_.*")
    interact_coef_map = {}
    for c in coefficients.index:
        if matcher.search(c):
            c_split = c.split("_")
            for j in c_split[2:]:
                interact_coef_map[(c_split[1], j)] = c
                if all((i == "x" for i in j)):  # wildcards also map to empty
                    interact_coef_map[(c_split[1], "")] = c

    for (cardinality, activity), coefs in interaction_coef.groupby(
        ["cardinality", "activity"]
    ):
        _logger.info(
            f"{n_persons} person households, interaction cardinality {cardinality}, activity {activity}"
        )
        if cardinality > n_persons:
            continue
        elif cardinality == n_persons:
            this_aname = activity * n_persons
            this_altnum = alts[this_aname]
            for rowindex, row in coefs.iterrows():
                expression = "&".join(
                    f"(p{p}_ptype == {t})"
                    for (p, t) in zip(person_numbers, row.interaction_ptypes)
                    if t != "*"
                )
                if expression:
                    if (activity, row.interaction_ptypes) in interact_coef_map:
                        linear_component = X(expression) * P(
                            interact_coef_map[(activity, row.interaction_ptypes)]
                        )
                    else:
                        linear_component = X(expression) * P(row.coefficient)
                else:
                    if (activity, row.interaction_ptypes) in interact_coef_map:
                        linear_component = P(
                            interact_coef_map[(activity, row.interaction_ptypes)]
                        )
                    else:
                        linear_component = P(row.coefficient)
                _logger.debug(
                    f"utility_co[{this_altnum} {this_aname}] += {linear_component}"
                )
                model.utility_co[this_altnum] += linear_component
        elif cardinality < n_persons:
            for combo in itertools.combinations(person_numbers, cardinality):
                pattern = interact_pattern(n_persons, combo, activity)
                for aname, anum in alts.items():
                    if pattern.match(aname):
                        for rowindex, row in coefs.iterrows():
                            expression = "&".join(
                                f"(p{p}_ptype == {t})"
                                for (p, t) in zip(combo, row.interaction_ptypes)
                                if t != "*"
                            )
                            # interaction terms without ptypes (i.e. with wildcards)
                            # only apply when the household size matches the cardinality
                            if expression != "":
                                if (
                                    activity,
                                    row.interaction_ptypes,
                                ) in interact_coef_map:
                                    linear_component = X(expression) * P(
                                        interact_coef_map[
                                            (activity, row.interaction_ptypes)
                                        ]
                                    )
                                else:
                                    linear_component = X(expression) * P(
                                        row.coefficient
                                    )
                                _logger.debug(
                                    f"utility_co[{anum} {aname}] += {linear_component}"
                                )
                                model.utility_co[anum] += linear_component


def cdap_split_data(households, values):
    if "cdap_rank" not in values:
        raise ValueError("assign cdap_rank to values first")
    # only process the first 5 household members
    values = values[values.cdap_rank <= 5]
    cdap_data = {}
    for hhsize, hhs_part in households.groupby(households.hhsize.clip(1, 5)):
        if hhsize == 1:
            v = pd.merge(values, hhs_part.household_id, on="household_id").set_index(
                "household_id"
            )
        else:
            v = (
                pd.merge(values, hhs_part.household_id, on="household_id")
                .set_index(["household_id", "cdap_rank"])
                .unstack()
            )
            v.columns = [f"p{i[1]}_{i[0]}" for i in v.columns]
            for agglom in ["override_choice", "model_choice"]:
                v[agglom] = v[[f"p{p}_{agglom}" for p in range(1, hhsize + 1)]].sum(1)
        cdap_data[hhsize] = v
    return cdap_data


def cdap_dataframes(households, values):
    data = cdap_split_data(households, values)
    dfs = {}
    for hhsize in data.keys():
        alts = generate_alternatives(hhsize)
        dfs[hhsize] = DataFrames(
            co=data[hhsize],
            alt_names=alts.keys(),
            alt_codes=alts.values(),
            av=1,
            ch=data[hhsize].override_choice.map(alts),
        )
    return dfs


# def _cdap_model(households, values, spec1, interaction_coef, coefficients):
#     cdap_data = cdap_dataframes(households, values)
#     m = {}
#     _logger.info(f"building for model 1")
#     m[1] = Model(dataservice=cdap_data[1])
#     cdap_base_utility_by_person(m[1], n_persons=1, spec=spec1)
#     m[1].choice_any = True
#     m[1].availability_any = True
#
#     # Add cardinality into interaction_coef if not present
#     if 'cardinality' not in interaction_coef:
#         interaction_coef['cardinality'] = interaction_coef['interaction_ptypes'].str.len()
#     for s in [2, 3, 4, 5]:
#         _logger.info(f"building for model {s}")
#         m[s] = Model(dataservice=cdap_data[s])
#         alts = generate_alternatives(s)
#         cdap_base_utility_by_person(m[s], s, spec1, alts, values.columns)
#         cdap_interaction_utility(m[s], s, alts, interaction_coef, coefficients)
#         m[s].choice_any = True
#         m[s].availability_any = True
#
#     result = ModelGroup(m.values())
#     explicit_value_parameters(result)
#     apply_coefficients(coefficients, result)
#     return result


def cdap_data(
    name="cdap",
    edb_directory="output/estimation_data_bundle/{name}/",
    coefficients_file="{name}_coefficients.csv",
    interaction_coeffs_file="{name}_interaction_coefficients.csv",
    households_file="../../final_households.csv",
    persons_file="../../final_persons.csv",
    spec1_file="{name}_INDIV_AND_HHSIZE1_SPEC.csv",
    settings_file="{name}_model_settings.yaml",
    chooser_data_file="{name}_values_combined.csv",
):
    edb_directory = edb_directory.format(name=name)
    if not os.path.exists(edb_directory):
        raise FileNotFoundError(edb_directory)

    def read_csv(filename, **kwargs):
        filename = filename.format(name=name)
        return pd.read_csv(os.path.join(edb_directory, filename), **kwargs)

    def read_yaml(filename, **kwargs):
        filename = filename.format(name=name)
        with open(os.path.join(edb_directory, filename), "rt") as f:
            return yaml.load(f, Loader=yaml.SafeLoader, **kwargs)

    settings = read_yaml(settings_file)

    try:
        hhs = read_csv(households_file)
    except FileNotFoundError:
        hhs = pd.read_csv(households_file)

    try:
        persons = read_csv(persons_file)
    except FileNotFoundError:
        persons = pd.read_csv(persons_file)

    person_type_map = settings.get("PERSON_TYPE_MAP")
    if person_type_map is None:
        raise KeyError("PERSON_TYPE_MAP missing from cdap_settings.yaml")

    person_rank = cdap.assign_cdap_rank(persons, person_type_map)

    coefficients = read_csv(
        coefficients_file,
        index_col="coefficient_name",
        comment="#",
    )

    interaction_coef = read_csv(
        interaction_coeffs_file,
        dtype={"interaction_ptypes": str},
        keep_default_na=False,
        comment="#",
    )

    spec1 = read_csv(spec1_file, comment="#")
    values = read_csv(chooser_data_file, comment="#")
    values["cdap_rank"] = person_rank

    return Dict(
        edb_directory=Path(edb_directory),
        person_data=values,
        spec1=spec1,
        interaction_coef=interaction_coef,
        coefficients=coefficients,
        households=hhs,
        settings=settings,
    )


def cdap_model(
    edb_directory="output/estimation_data_bundle/{name}/",
    coefficients_file="{name}_coefficients.csv",
    interaction_coeffs_file="{name}_interaction_coefficients.csv",
    households_file="../../final_households.csv",
    persons_file="../../final_persons.csv",
    spec1_file="{name}_INDIV_AND_HHSIZE1_SPEC.csv",
    settings_file="{name}_model_settings.yaml",
    chooser_data_file="{name}_values_combined.csv",
    return_data=False,
):
    d = cdap_data(
        name="cdap",
        edb_directory=edb_directory,
        coefficients_file=coefficients_file,
        interaction_coeffs_file=interaction_coeffs_file,
        households_file=households_file,
        persons_file=persons_file,
        spec1_file=spec1_file,
        settings_file=settings_file,
        chooser_data_file=chooser_data_file,
    )

    households = d.households
    values = d.person_data
    spec1 = d.spec1
    interaction_coef = d.interaction_coef
    coefficients = d.coefficients

    cdap_dfs = cdap_dataframes(households, values)
    m = {}
    _logger.info(f"building for model 1")
    m[1] = Model(dataservice=cdap_dfs[1])
    cdap_base_utility_by_person(m[1], n_persons=1, spec=spec1)
    m[1].choice_any = True
    m[1].availability_any = True

    # Add cardinality into interaction_coef if not present
    if "cardinality" not in interaction_coef:
        interaction_coef["cardinality"] = interaction_coef[
            "interaction_ptypes"
        ].str.len()
    for s in [2, 3, 4, 5]:
        _logger.info(f"building for model {s}")
        m[s] = Model(dataservice=cdap_dfs[s])
        alts = generate_alternatives(s)
        cdap_base_utility_by_person(m[s], s, spec1, alts, values.columns)
        cdap_interaction_utility(m[s], s, alts, interaction_coef, coefficients)
        m[s].choice_any = True
        m[s].availability_any = True

    model = ModelGroup(m.values())
    explicit_value_parameters(model)
    apply_coefficients(coefficients, model)
    if return_data:
        return model, d
    return model



# ActivitySim
# See full license in LICENSE.txt.

import logging
import os
import sys

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
logger.addHandler(ch)


inputs = {
    "households": "final_households.csv",
    "persons": "final_persons.csv",
    "tours": "final_tours.csv",
    "joint_tour_participants": "final_joint_tour_participants.csv",
    "trips": "final_trips.csv",
}

surveys = {
    "households": "survey_households.csv",
    "persons": "survey_persons.csv",
    "tours": "survey_tours.csv",
    "joint_tour_participants": "survey_joint_tour_participants.csv",
    "trips": "survey_trips.csv",
}


args = sys.argv[1:]
assert len(args) == 1, "usage extract_survey_data.py <data_dir>"

data_dir = args[0]

input_dir = os.path.join(data_dir, "survey_data/")
output_dir = os.path.join(data_dir, "survey_data/")

configs_dir = os.path.dirname("../example/configs/")

households = pd.read_csv(os.path.join(input_dir, inputs["households"]))
persons = pd.read_csv(os.path.join(input_dir, inputs["persons"]))
tours = pd.read_csv(os.path.join(input_dir, inputs["tours"]))
joint_tour_participants = pd.read_csv(
    os.path.join(input_dir, inputs["joint_tour_participants"])
)
trips = pd.read_csv(os.path.join(input_dir, inputs["trips"]))

households = households[
    [
        "household_id",
        "home_zone_id",
        "income",
        "hhsize",
        "HHT",
        "auto_ownership",
        "num_workers",
    ]
]
persons = persons[
    [
        "person_id",
        "household_id",
        "age",
        "PNUM",
        "sex",
        "pemploy",
        "pstudent",
        "ptype",
        "school_zone_id",
        "workplace_zone_id",
        "free_parking_at_work",
    ]
]
tours = tours[
    [
        "tour_id",
        "person_id",
        "household_id",
        "tour_type",
        "tour_category",
        "destination",
        "origin",
        "start",
        "end",
        "tour_mode",
        "parent_tour_id",
    ]
]
joint_tour_participants = joint_tour_participants[
    ["participant_id", "tour_id", "household_id", "person_id", "participant_num"]
]

OPTIONAL_TRIP_COLUMNS = []
# FIXME trip order is ambiguous if two trips have same start time?
# OPTIONAL_TRIP_COLUMNS = ['trip_num']

trips = trips[
    [
        "trip_id",
        "person_id",
        "household_id",
        "tour_id",
        "outbound",
        "purpose",
        "destination",
        "origin",
        "depart",
        "trip_mode",
    ]
    + OPTIONAL_TRIP_COLUMNS
]

households.to_csv(os.path.join(output_dir, surveys["households"]), index=False)
persons.to_csv(os.path.join(output_dir, surveys["persons"]), index=False)
tours.to_csv(os.path.join(output_dir, surveys["tours"]), index=False)
joint_tour_participants.to_csv(
    os.path.join(output_dir, surveys["joint_tour_participants"]), index=False
)
trips.to_csv(os.path.join(output_dir, surveys["trips"]), index=False)



# ActivitySim
# See full license in LICENSE.txt.


import os

import pandas as pd
import pandas.testing as pdt
import pytest

from ..tour_frequency import process_mandatory_tours


def mandatory_tour_frequency_alternatives():
    configs_dir = os.path.join(os.path.dirname(__file__), "configs")
    f = os.path.join(configs_dir, "mandatory_tour_frequency_alternatives.csv")
    df = pd.read_csv(f, comment="#")
    df.set_index("alt", inplace=True)
    return df


def test_mtf():
    persons = pd.DataFrame(
        {
            "is_worker": [True, True, False, False],
            "mandatory_tour_frequency": [
                "work1",
                "work_and_school",
                "work_and_school",
                "school2",
            ],
            "school_zone_id": [1, 2, 3, 4],
            "workplace_zone_id": [10, 20, 30, 40],
            "home_zone_id": [100, 200, 300, 400],
            "household_id": [1, 2, 2, 4],
        },
        index=[10, 20, 30, 40],
    )

    tour_frequency_alternatives = mandatory_tour_frequency_alternatives()

    mandatory_tours = process_mandatory_tours(persons, tour_frequency_alternatives)

    idx = mandatory_tours.index

    pdt.assert_series_equal(
        mandatory_tours.person_id,
        pd.Series([10, 20, 20, 30, 30, 40, 40], index=idx, name="person_id"),
    )

    pdt.assert_series_equal(
        mandatory_tours.tour_type,
        pd.Series(
            ["work", "work", "school", "work", "school", "school", "school"],
            index=idx,
            name="tour_type",
        ),
    )

    # tour_nums for work_and_school non-worker should be flipped
    pdt.assert_series_equal(
        mandatory_tours.tour_num,
        pd.Series([1, 1, 2, 2, 1, 1, 2], index=idx, name="tour_num"),
    )

    pdt.assert_series_equal(
        mandatory_tours.destination,
        pd.Series([10, 20, 2, 30, 3, 4, 4], index=idx, name="destination"),
    )

    pdt.assert_series_equal(
        mandatory_tours.origin,
        pd.Series([100, 200, 200, 300, 300, 400, 400], index=idx, name="origin"),
    )



# ActivitySim
# See full license in LICENSE.txt.

import logging
import os
import shutil

import pandas as pd
import yaml

from activitysim.abm.models.util import canonical_ids as cid
from activitysim.core import config, simulate
from activitysim.core.util import reindex

logger = logging.getLogger("estimation")

ESTIMATION_SETTINGS_FILE_NAME = "estimation.yaml"


def unlink_files(directory_path, file_types=("csv", "yaml")):
    for file_name in os.listdir(directory_path):
        if file_name.endswith(file_types):
            file_path = os.path.join(directory_path, file_name)
            try:
                if os.path.isfile(file_path):
                    os.unlink(file_path)
                    print(f"deleted {file_path}")
            except Exception as e:
                print(e)


class Estimator(object):
    def __init__(self, bundle_name, model_name, estimation_table_recipes):

        logger.info("Initialize Estimator for'%s'" % (model_name,))

        self.bundle_name = bundle_name
        self.model_name = model_name
        self.settings_name = model_name
        self.estimation_table_recipes = estimation_table_recipes
        self.estimating = True

        # ensure the output data directory exists
        output_dir = self.output_directory()
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)  # make directory if needed

        # delete estimation files
        unlink_files(self.output_directory(), file_types=("csv", "yaml"))
        if self.bundle_name != self.model_name:
            # kind of inelegant to always delete these, but ok as they are redundantly recreated for each sub model
            unlink_files(
                self.output_directory(bundle_directory=True), file_types=("csv", "yaml")
            )

        # FIXME - not required?
        # assert 'override_choices' in self.model_settings, \
        #     "override_choices not found for %s in %s." % (model_name, ESTIMATION_SETTINGS_FILE_NAME)

        self.omnibus_tables = self.estimation_table_recipes["omnibus_tables"]
        self.omnibus_tables_append_columns = self.estimation_table_recipes[
            "omnibus_tables_append_columns"
        ]
        self.tables = {}
        self.tables_to_cache = [
            table_name
            for tables in self.omnibus_tables.values()
            for table_name in tables
        ]
        self.alt_id_column_name = None
        self.chooser_id_column_name = None

    @property
    def want_unsampled_alternatives(self):
        # use complete set of alternatives for estimation of interaction simulation (e.g. destination choice)
        # WARNING location_choice expects unsample alts so it can retrieve mode choice logsums for overridden choice
        # if we allow estimation based on sampled alternatives, location_choice may need to compute logsum
        return True

    def log(self, msg, level=logging.INFO):
        logger.log(level, "%s: %s" % (self.model_name, msg))

    def info(self, msg):
        self.log(msg, level=logging.INFO)

    def debug(self, msg):
        self.log(msg, level=logging.DEBUG)

    def warning(self, msg):
        self.log(msg, level=logging.WARNING)

    def set_alt_id(self, alt_id):
        self.alt_id_column_name = alt_id

    def get_alt_id(self):
        if self.alt_id_column_name is None:
            self.warning("alt_id is None. Did you forget to call set_alt_id()?")
        assert self.alt_id_column_name is not None
        return self.alt_id_column_name

    def set_chooser_id(self, chooser_id_column_name):
        self.chooser_id_column_name = chooser_id_column_name

    def get_chooser_id(self):
        if self.chooser_id_column_name is None:
            self.warning("chooser_id is None. Did you forget to call set_chooser_id()?")
        assert self.chooser_id_column_name is not None
        return self.chooser_id_column_name

    def end_estimation(self):

        self.write_omnibus_table()

        self.estimating = False
        self.tables = None

        self.info("end estimation")

        manager.release(self)

    def output_directory(self, bundle_directory=False):

        # shouldn't be asking for this if not estimating
        assert self.estimating
        assert self.model_name is not None

        dir = os.path.join(
            config.output_file_path("estimation_data_bundle"), self.bundle_name
        )

        if bundle_directory:
            # shouldn't be asking - probably confused
            assert self.bundle_name != self.model_name

        if self.bundle_name != self.model_name and not bundle_directory:
            dir = os.path.join(dir, self.model_name)

        return dir

    def output_file_path(self, table_name, file_type=None, bundle_directory=False):

        # shouldn't be asking for this if not estimating
        assert self.estimating

        output_dir = self.output_directory(bundle_directory)

        if bundle_directory:
            file_name = f"{self.bundle_name}_{table_name}"
        else:
            if self.model_name == self.bundle_name:
                file_name = f"{self.model_name}_{table_name}"
            else:
                file_name = f"{self.bundle_name}_{table_name}"

        if file_type and os.path.splitext(file_name)[1] != f".{file_type}":
            file_name = f"{file_name}.{file_type}"

        return os.path.join(output_dir, file_name)

    def write_table(
        self, df, table_name, index=True, append=True, bundle_directory=False
    ):
        """

        Parameters
        ----------
        df
        table_name: str
            if table_name has file type, then pass through filename without adding model or bundle name prefix
        index: booelan
        append: boolean
        bundle_directory: boolean

        """

        def cache_table(df, table_name, append):
            if table_name in self.tables and not append:
                raise RuntimeError(
                    "cache_table %s append=False and table exists" % (table_name,)
                )
            if table_name in self.tables:
                self.tables[table_name] = pd.concat([self.tables[table_name], df])
            else:
                self.tables[table_name] = df.copy()

        def write_table(df, table_name, index, append, bundle_directory):
            if table_name.endswith(".csv"):
                # pass through filename without adding model or bundle name prefix
                file_path = os.path.join(
                    self.output_directory(bundle_directory), table_name
                )
            else:
                file_path = self.output_file_path(table_name, "csv", bundle_directory)
            file_exists = os.path.isfile(file_path)
            if file_exists and not append:
                raise RuntimeError(
                    "write_table %s append=False and file exists: %s"
                    % (table_name, file_path)
                )
            df.to_csv(file_path, mode="a", index=index, header=(not file_exists))

        assert self.estimating

        cache = table_name in self.tables_to_cache
        write = not cache
        # write = True

        if cache:
            cache_table(df, table_name, append)
            self.debug("write_table cache: %s" % table_name)

        if write:
            write_table(df, table_name, index, append, bundle_directory)
            self.debug("write_table write: %s" % table_name)

    def write_omnibus_table(self):

        if len(self.omnibus_tables) == 0:
            return

        for omnibus_table, table_names in self.omnibus_tables.items():

            self.debug(
                "write_omnibus_table: %s table_names: %s" % (omnibus_table, table_names)
            )
            for t in table_names:
                if t not in self.tables:
                    self.warning(
                        "write_omnibus_table: %s table '%s' not found"
                        % (omnibus_table, t)
                    )

            # ignore any tables not in cache
            table_names = [t for t in table_names if t in self.tables]
            concat_axis = (
                1 if omnibus_table in self.omnibus_tables_append_columns else 0
            )

            df = pd.concat([self.tables[t] for t in table_names], axis=concat_axis)
            df.sort_index(ascending=True, inplace=True, kind="mergesort")

            file_path = self.output_file_path(omnibus_table, "csv")
            assert not os.path.isfile(file_path)

            df.to_csv(file_path, mode="a", index=True, header=True)

            self.debug("write_omnibus_choosers: %s" % file_path)

    def write_dict(self, d, dict_name, bundle_directory):

        assert self.estimating

        file_path = self.output_file_path(dict_name, "yaml", bundle_directory)

        # we don't know how to concat, and afraid to overwrite
        assert not os.path.isfile(file_path)

        with open(file_path, "w") as f:
            # write ordered dict as array
            yaml.dump(d, f)

        self.debug("estimate.write_dict: %s" % file_path)

    def write_coefficients(
        self, coefficients_df=None, model_settings=None, file_name=None
    ):
        """
        Because the whole point of estimation is to generate new coefficient values
        we want to make it easy to put the coefficients file back in configs
        So we make a point of preserving the same filename as the original config file
        """

        if model_settings is not None:
            assert file_name is None
            file_name = model_settings["COEFFICIENTS"]

        assert file_name is not None

        if coefficients_df is None:
            coefficients_df = simulate.read_model_coefficients(file_name=file_name)

        # preserve original config file name
        base_file_name = os.path.basename(file_name)

        assert self.estimating
        self.write_table(coefficients_df, base_file_name, append=False)

    def write_coefficients_template(self, model_settings):
        assert self.estimating

        coefficients_df = simulate.read_model_coefficient_template(model_settings)
        tag = "coefficients_template"
        self.write_table(coefficients_df, tag, append=False)

    def write_choosers(self, choosers_df):
        self.write_table(choosers_df, "choosers", append=True)

    def write_choices(self, choices):
        if isinstance(choices, pd.Series):
            choices = choices.to_frame(name="model_choice")
        assert list(choices.columns) == ["model_choice"]
        self.write_table(choices, "choices", append=True)

    def write_override_choices(self, choices):
        if isinstance(choices, pd.Series):
            choices = choices.to_frame(name="override_choice")
        assert list(choices.columns) == ["override_choice"]
        self.write_table(choices, "override_choices", append=True)

    def write_constants(self, constants):
        self.write_dict(self, constants, "model_constants")

    def write_nest_spec(self, nest_spec):
        self.write_dict(self, nest_spec, "nest_spec")

    def copy_model_settings(
        self, settings_file_name, tag="model_settings", bundle_directory=False
    ):

        input_path = config.base_settings_file_path(settings_file_name)

        output_path = self.output_file_path(tag, "yaml", bundle_directory)

        shutil.copy(input_path, output_path)

    def write_model_settings(
        self, model_settings, settings_file_name, bundle_directory=False
    ):

        if "include_settings" in model_settings:
            file_path = self.output_file_path(
                "model_settings", "yaml", bundle_directory
            )
            assert not os.path.isfile(file_path)
            with open(file_path, "w") as f:
                yaml.dump(model_settings, f)
        else:
            self.copy_model_settings(
                settings_file_name, bundle_directory=bundle_directory
            )
        if "inherit_settings" in model_settings:
            self.write_dict(
                model_settings, "inherited_model_settings", bundle_directory
            )

    def melt_alternatives(self, df):

        alt_id_name = self.alt_id_column_name

        assert alt_id_name is not None, (
            "alt_id not set. Did you forget to call set_alt_id()? (%s)"
            % self.model_name
        )
        assert (
            alt_id_name in df
        ), "alt_id_column_name '%s' not in alternatives table (%s)" % (
            alt_id_name,
            self.model_name,
        )

        variable_column = "variable"

        #            alt_dest  util_dist_0_1  util_dist_1_2  ...
        # person_id                                          ...
        # 31153             1            1.0           0.75  ...
        # 31153             2            1.0           0.46  ...
        # 31153             3            1.0           0.28  ...

        if df.index.name is not None:
            chooser_name = df.index.name
            assert self.chooser_id_column_name in (chooser_name, None)
            df = df.reset_index()
        else:
            assert self.chooser_id_column_name is not None
            chooser_name = self.chooser_id_column_name
            assert chooser_name in df

        # mergesort is the only stable sort, and we want the expressions to appear in original df column order
        melt_df = (
            pd.melt(df, id_vars=[chooser_name, alt_id_name])
            .sort_values(by=chooser_name, kind="mergesort")
            .rename(columns={"variable": variable_column})
        )

        # person_id,alt_dest,expression,value
        # 31153,1,util_dist_0_1,1.0
        # 31153,2,util_dist_0_1,1.0
        # 31153,3,util_dist_0_1,1.0

        melt_df = melt_df.set_index(
            [chooser_name, variable_column, alt_id_name]
        ).unstack(2)
        melt_df.columns = melt_df.columns.droplevel(0)
        melt_df = melt_df.reset_index(1)

        # person_id,expression,1,2,3,4,5,...
        # 31153,util_dist_0_1,0.75,0.46,0.27,0.63,0.48,...
        # 31153,util_dist_1_2,0.0,0.0,0.0,0.0,0.0,...
        # 31153,util_dist_2_3,0.0,0.0,0.0,0.0,0.0,...

        return melt_df

    def write_interaction_expression_values(self, df):
        df = self.melt_alternatives(df)
        self.write_table(df, "interaction_expression_values", append=True)

    def write_expression_values(self, df):
        self.write_table(df, "expression_values", append=True)

    def write_alternatives(self, alternatives_df, bundle_directory=False):
        self.write_table(
            alternatives_df,
            "alternatives",
            append=True,
            bundle_directory=bundle_directory,
        )

    def write_interaction_sample_alternatives(self, alternatives_df):
        alternatives_df = self.melt_alternatives(alternatives_df)
        self.write_table(
            alternatives_df, "interaction_sample_alternatives", append=True
        )

    def write_interaction_simulate_alternatives(self, interaction_df):
        interaction_df = self.melt_alternatives(interaction_df)
        self.write_table(
            interaction_df, "interaction_simulate_alternatives", append=True
        )

    def get_survey_values(self, model_values, table_name, column_names):
        # convenience method so deep callers don't need to import estimation
        assert self.estimating
        return manager.get_survey_values(model_values, table_name, column_names)

    def get_survey_table(self, table_name):
        # convenience method so deep callers don't need to import estimation
        assert self.estimating
        return manager.get_survey_table(table_name)

    def write_spec(
        self, model_settings=None, file_name=None, tag="SPEC", bundle_directory=False
    ):

        if model_settings is not None:
            assert file_name is None
            file_name = model_settings[tag]

        input_path = config.config_file_path(file_name)

        table_name = tag  # more readable than full spec file_name
        output_path = self.output_file_path(table_name, "csv", bundle_directory)
        shutil.copy(input_path, output_path)
        self.debug("estimate.write_spec: %s" % output_path)


class EstimationManager(object):
    def __init__(self):

        self.settings_initialized = False
        self.bundles = []
        self.estimation_table_recipes = {}
        self.model_estimation_table_types = {}
        self.estimating = {}

    def initialize_settings(self):

        # FIXME - can't we just initialize in init and handle no-presence of settings file as not enabled
        if self.settings_initialized:
            return

        assert not self.settings_initialized
        settings = config.read_model_settings(ESTIMATION_SETTINGS_FILE_NAME)
        self.enabled = settings.get("enable", "True")
        self.bundles = settings.get("bundles", [])

        self.model_estimation_table_types = settings.get(
            "model_estimation_table_types", {}
        )
        self.estimation_table_recipes = settings.get("estimation_table_recipes", {})

        if self.enabled:

            self.survey_tables = settings.get("survey_tables", {})
            for table_name, table_info in self.survey_tables.items():
                assert (
                    "file_name" in table_info
                ), "No file name specified for survey_table '%s' in %s" % (
                    table_name,
                    ESTIMATION_SETTINGS_FILE_NAME,
                )
                file_path = config.data_file_path(
                    table_info["file_name"], mandatory=True
                )
                assert os.path.exists(
                    file_path
                ), "File for survey table '%s' not found: %s" % (table_name, file_path)
                df = pd.read_csv(file_path)
                index_col = table_info.get("index_col")
                if index_col is not None:
                    assert (
                        index_col in df.columns
                    ), "Index col '%s' not in survey_table '%s' in file: %s % (index_col, table_name, file_path)"
                    df.set_index(index_col, inplace=True)

                # add the table df to survey_tables
                table_info["df"] = df

        self.settings_initialized = True

    def begin_estimation(self, model_name, bundle_name=None):
        """
        begin estimating of model_name is specified as model to estimate, otherwise return False

        Parameters
        ----------
        model_name

        Returns
        -------

        """
        # load estimation settings file
        if not self.settings_initialized:
            self.initialize_settings()

        # global estimation setting
        if not self.enabled:
            return None

        bundle_name = bundle_name or model_name

        if bundle_name not in self.bundles:
            logger.warning(
                f"estimation bundle {bundle_name} not in settings file {ESTIMATION_SETTINGS_FILE_NAME}"
            )
            return None

        # can't estimate the same model simultaneously
        assert (
            model_name not in self.estimating
        ), "Cant begin estimating %s - already estimating that model." % (model_name,)

        assert (
            bundle_name in self.model_estimation_table_types
        ), "No estimation_table_type for %s in %s." % (
            bundle_name,
            ESTIMATION_SETTINGS_FILE_NAME,
        )

        model_estimation_table_type = self.model_estimation_table_types[bundle_name]

        assert (
            model_estimation_table_type in self.estimation_table_recipes
        ), "model_estimation_table_type '%s' for model %s no in %s." % (
            model_estimation_table_type,
            model_name,
            ESTIMATION_SETTINGS_FILE_NAME,
        )

        self.estimating[model_name] = Estimator(
            bundle_name,
            model_name,
            estimation_table_recipes=self.estimation_table_recipes[
                model_estimation_table_type
            ],
        )

        return self.estimating[model_name]

    def release(self, estimator):

        self.estimating.pop(estimator.model_name)

    def get_survey_table(self, table_name):
        assert self.enabled
        if table_name not in self.survey_tables:
            logger.warning(
                "EstimationManager. get_survey_table: survey table '%s' not in survey_tables"
                % table_name
            )
        df = self.survey_tables[table_name].get("df")
        return df

    def get_survey_values(self, model_values, table_name, column_names):

        assert isinstance(
            model_values, (pd.Series, pd.DataFrame, pd.Index)
        ), "get_survey_values model_values has unrecognized type %s" % type(
            model_values
        )

        dest_index = (
            model_values if isinstance(model_values, (pd.Index)) else model_values.index
        )

        # read override_df table
        survey_df = manager.get_survey_table(table_name)

        assert survey_df is not None, "get_survey_values: table '%s' not found" % (
            table_name,
        )

        column_name = column_names if isinstance(column_names, str) else None
        if column_name:
            column_names = [column_name]

        if not set(column_names).issubset(set(survey_df.columns)):
            missing_columns = list(set(column_names) - set(survey_df.columns))
            logger.error(
                "missing columns (%s) in survey table %s"
                % (missing_columns, table_name)
            )
            print("survey table columns: %s" % (survey_df.columns,))
            raise RuntimeError(
                "missing columns (%s) in survey table %s"
                % (missing_columns, table_name)
            )

        assert set(column_names).issubset(set(survey_df.columns)), (
            f"missing columns ({list(set(column_names) - set(survey_df.columns))}) "
            f"in survey table {table_name} {list(survey_df.columns)}"
        )

        # for now tour_id is asim_tour_id in survey_df
        asim_df_index_name = dest_index.name
        if asim_df_index_name == survey_df.index.name:
            # survey table has same index as activitysim
            survey_df_index_column = "index"
        elif asim_df_index_name in survey_df.columns:
            # survey table has activitysim index as column
            survey_df_index_column = asim_df_index_name
        elif "asim_%s" % asim_df_index_name in survey_df.columns:
            # survey table has activitysim index as column with asim_ prefix
            survey_df_index_column = "asim_%s" % asim_df_index_name
        else:
            logger.error(
                "get_survey_values:index '%s' not in survey table" % dest_index.name
            )
            # raise RuntimeError("index '%s' not in survey table %s" % (dest_index.name, table_name)
            survey_df_index_column = None

        logger.debug(
            "get_survey_values: reindexing using %s.%s"
            % (table_name, survey_df_index_column)
        )

        values = pd.DataFrame(index=dest_index)
        for c in column_names:
            if survey_df_index_column == "index":
                survey_values = survey_df[c]
            else:
                survey_values = pd.Series(
                    survey_df[c].values, index=survey_df[survey_df_index_column]
                )

            survey_values = reindex(survey_values, dest_index)

            # shouldn't be any choices we can't override
            missing_values = survey_values.isna()
            if missing_values.any():
                logger.error(
                    "missing survey_values for %s\n%s" % (c, dest_index[missing_values])
                )
                logger.error(
                    "couldn't get_survey_values for %s in %s\n" % (c, table_name)
                )
                raise RuntimeError(
                    "couldn't get_survey_values for %s in %s\n" % (c, table_name)
                )

            values[c] = survey_values

        return values[column_name] if column_name else values


manager = EstimationManager()



# remove some of the asim-style columns added by marin_work_tour_mode_choice.py
# so data input files look 'realistic' - and that work is done instaed by 'import_tours' annotation expression files

import os

import openmatrix as omx
import pandas as pd

input_dir = "./data_3_marin"
output_dir = (
    "./data_3_marin/fix"  # don't overwrite - but these files shold replace 'oritinals'
)


def input_path(filenane):
    return os.path.join(input_dir, filenane)


def output_path(filenane):
    return os.path.join(output_dir, filenane)


# 0 - get county zones

mazs = pd.read_csv(input_path("maz_data_asim.csv"))
del mazs["zone_id"]
del mazs["county_id"]
mazs.to_csv(output_path("maz_data_asim.csv"), index=False)

tazs = mazs["TAZ"].unique()
tazs.sort()
assert ((tazs - 1) == range(len(tazs))).all()

# MAZ,TAZ
taps = pd.read_csv(input_path("maz_taz.csv"))
# nothing
taps.to_csv(output_path("maz_taz.csv"), index=False)

taps = pd.read_csv(input_path("tap_data.csv"))
# nothing
taps.to_csv(output_path("tap_data.csv"), index=False)


# 2 - nearby skims need headers

maz_tap_walk = pd.read_csv(input_path("maz_tap_walk.csv"))
maz_maz_walk = pd.read_csv(input_path("maz_maz_walk.csv"))
maz_maz_bike = pd.read_csv(input_path("maz_maz_bike.csv"))

del maz_tap_walk["TAP.1"]
del maz_maz_walk["DMAZ.1"]
del maz_maz_bike["DMAZ.1"]

maz_tap_walk.to_csv(output_path("maz_tap_walk.csv"), index=False)
maz_maz_walk.to_csv(output_path("maz_maz_walk.csv"), index=False)
maz_maz_bike.to_csv(output_path("maz_maz_bike.csv"), index=False)

# 3 - accessibility data

access = pd.read_csv(input_path("access.csv"))
del access["zone_id"]
access.to_csv(output_path("access.csv"), index=False)

# 4 - maz to tap drive data

taz_tap_drive = pd.read_csv(input_path("maz_taz_tap_drive.csv"))

taz_tap_drive.to_csv(output_path("maz_taz_tap_drive.csv"), index=False)

# 5 - households

households = pd.read_csv(input_path("households_asim.csv"))
del households["home_zone_id"]
del households["household_id"]

households.to_csv(output_path("households_asim.csv"), index=False)

# 6 - persons

persons = pd.read_csv(input_path("persons_asim.csv"))
del persons["person_id"]
del persons["household_id"]
del persons["is_university"]
persons.to_csv(output_path("persons_asim.csv"), index=False)

# 7 - tours file

work_tours = pd.read_csv(input_path("work_tours.csv"))
del work_tours["household_id"]
del work_tours["destination"]
del work_tours["start"]
del work_tours["end"]
del work_tours["tour_type"]
work_tours.to_csv(output_path("work_tours.csv"), index=False)



import json
import os
import shutil
import subprocess
import sys

ASV_CONFIG = {
    # The version of the config file format.  Do not change, unless
    # you know what you are doing.
    "version": 1,
    # The name of the project being benchmarked
    "project": "activitysim",
    # The project's homepage
    "project_url": "https://activitysim.github.io/",
    # The URL or local path of the source code repository for the
    # project being benchmarked
    "repo": ".",
    # The tool to use to create environments.
    "environment_type": "conda",
    # the base URL to show a commit for the project.
    "show_commit_url": "http://github.com/ActivitySim/activitysim/commit/",
    # The Pythons you'd like to test against.  If not provided, defaults
    # to the current version of Python used to run `asv`.
    # "pythons": ["2.7", "3.6"],
    # The list of conda channel names to be searched for benchmark
    # dependency packages in the specified order
    "conda_channels": ["conda-forge"],
    # The matrix of dependencies to test.  Each key is the name of a
    # package (in PyPI) and the values are version numbers.  An empty
    # list or empty string indicates to just test against the default
    # (latest) version. null indicates that the package is to not be
    # installed. If the package to be tested is only available from
    # PyPi, and the 'environment_type' is conda, then you can preface
    # the package name by 'pip+', and the package will be installed via
    # pip (with all the conda available packages installed first,
    # followed by the pip installed packages).
    "matrix": {
        "pyarrow": [],
        "numpy": [],
        "openmatrix": [],
        "pandas": [],
        "pyyaml": [],
        "pytables": [],
        "toolz": [],
        "orca": [],
        "psutil": [],
        "requests": [],
        "numba": [],
        "coverage": [],
        "pytest": [],
        "cytoolz": [],
    },
    # The directory (relative to the current directory) to cache the Python
    # environments in.  If not provided, defaults to "env"
    # "env_dir": "../activitysim-asv/env",
    # The directory (relative to the current directory) that raw benchmark
    # results are stored in.  If not provided, defaults to "results".
    # "results_dir": "../activitysim-asv/results",
    # The directory (relative to the current directory) that the html tree
    # should be written to.  If not provided, defaults to "html".
    # "html_dir": "../activitysim-asv/html",
    # List of branches to benchmark. If not provided, defaults to "master"
    # (for git) or "default" (for mercurial).
    "branches": ["develop"],
}


def make_asv_argparser(parser):
    """
    The entry point for asv.

    Most of this work is handed off to the airspeed velocity library.
    """
    try:
        from asv.commands import Command, command_order, common_args, util
    except ImportError:
        return

    def help(args):
        parser.print_help()
        sys.exit(0)

    common_args.add_global_arguments(parser, suppress_defaults=False)

    subparsers = parser.add_subparsers(
        title="benchmarking with airspeed velocity", description="valid subcommands"
    )

    help_parser = subparsers.add_parser("help", help="Display usage information")
    help_parser.set_defaults(afunc=help)

    commands = dict((x.__name__, x) for x in util.iter_subclasses(Command))

    hide_commands = [
        "quickstart",
    ]

    for command in command_order:
        if str(command) in hide_commands:
            continue
        subparser = commands[str(command)].setup_arguments(subparsers)
        common_args.add_global_arguments(subparser)
        subparser.add_argument(
            "--workspace",
            "-w",
            help="benchmarking workspace directory",
            default=".",
        )
        subparser.add_argument(
            "--branch",
            type=str,
            action="append",
            metavar="NAME",
            help="git branch to include in benchmarking",
        )
        del commands[command]

    for name, command in sorted(commands.items()):
        if str(command) in hide_commands:
            continue
        subparser = command.setup_arguments(subparsers)
        subparser.add_argument(
            "--workspace",
            "-w",
            help="benchmarking workspace directory",
            default=".",
        )
        subparser.add_argument(
            "--branch",
            type=str,
            action="append",
            metavar="NAME",
            help="git branch to include in benchmarking",
        )
        common_args.add_global_arguments(subparser)

    from ..benchmarking.latest import Batch, Latest

    subparser = Latest.setup_arguments(subparsers)
    subparser.add_argument(
        "--workspace",
        "-w",
        help="benchmarking workspace directory",
        default=".",
    )
    subparser.add_argument(
        "--branch",
        type=str,
        action="append",
        metavar="NAME",
        help="git branch to include in benchmarking",
    )
    common_args.add_global_arguments(subparser)

    subparser = Batch.setup_arguments(subparsers)
    subparser.add_argument(
        "--workspace",
        "-w",
        help="benchmarking workspace directory",
        default=".",
    )

    from ..benchmarking.profile_inspector import ProfileInspector

    subparser = ProfileInspector.setup_arguments(subparsers)
    subparser.add_argument(
        "--workspace",
        "-w",
        help="benchmarking workspace directory",
        default=".",
    )

    parser.set_defaults(afunc=benchmark)
    return parser, subparsers


def benchmark(args):
    try:
        import asv
    except ModuleNotFoundError:
        print("airspeed velocity is not installed")
        print("try `conda install asv -c conda-forge` if you want to run benchmarks")
        sys.exit(1)
    from asv import util
    from asv.console import log

    log.enable(args.verbose)

    log.info("<== benchmarking activitysim ==>")

    # workspace
    args.workspace = os.path.abspath(args.workspace)

    if os.path.abspath(os.path.expanduser("~")) == args.workspace:
        log.error(
            "don't run benchmarks in the user's home directory \n"
            "try changing directories before calling `activitysim benchmark` "
            "or use the --workspace option \n"
        )
        sys.exit(1)

    if not os.path.isdir(args.workspace):
        raise NotADirectoryError(args.workspace)
    log.info(f" workspace: {args.workspace}")
    os.chdir(args.workspace)
    os.environ["ASIM_ASV_WORKSPACE"] = str(args.workspace)

    from ..benchmarking import workspace

    workspace.set_dir(args.workspace)

    from .. import __path__ as pkg_path

    log.info(f" activitysim installation: {pkg_path[0]}")

    repo_dir = os.path.normpath(os.path.join(pkg_path[0], ".."))
    git_dir = os.path.normpath(os.path.join(repo_dir, ".git"))
    local_git = os.path.exists(git_dir)
    log.info(f" local git repo available: {local_git}")

    branches = args.branch

    asv_config = ASV_CONFIG.copy()
    if local_git:
        repo_dir_rel = os.path.relpath(repo_dir, args.workspace)
        log.info(f" local git repo: {repo_dir_rel}")
        asv_config["repo"] = repo_dir_rel
        if not branches:
            # add current branch to the branches to benchmark
            current_branch = (
                subprocess.check_output(
                    ["git", "branch", "--show-current"],
                    env={"GIT_DIR": git_dir},
                    stdin=None,
                    stderr=None,
                    shell=False,
                    universal_newlines=False,
                )
                .decode()
                .strip()
            )
            if current_branch:
                asv_config["branches"].append(current_branch)
    else:
        log.info(f" local git repo available: {local_git}")
        asv_config["repo"] = "https://github.com/ActivitySim/activitysim.git"

    asv_config["branches"].extend(branches)

    # copy the benchmarks to the workspace, deleting previous files in workspace
    import activitysim.benchmarking.benchmarks

    benchmarks_dir = os.path.dirname(activitysim.benchmarking.benchmarks.__file__)
    shutil.rmtree(
        os.path.join(args.workspace, "benchmarks"),
        ignore_errors=True,
    )
    shutil.copytree(
        benchmarks_dir,
        os.path.join(args.workspace, "benchmarks"),
        dirs_exist_ok=True,
    )

    # write the asv config to the workspace
    conf_file = os.path.normpath(os.path.join(args.workspace, "asv.conf.json"))
    with open(conf_file, "wt") as jf:
        json.dump(asv_config, jf)

    if args.config and args.config != "asv.conf.json":
        raise ValueError(
            "activitysim manages the asv config json file itself, do not use --config"
        )
    args.config = os.path.abspath(conf_file)

    # write the pre-commit search and replace hook to the workspace
    search_replace_file = os.path.normpath(
        os.path.join(args.workspace, ".pre-commit-search-and-replace.yaml")
    )
    with open(search_replace_file, "wt") as sf:
        benchpath = os.path.join(args.workspace, "benchmarks")
        if not benchpath.endswith(os.path.sep):
            benchpath += os.path.sep
        benchpath = benchpath.replace(os.path.sep, r"[/\\]")
        sf.write(f"""- search: /{benchpath}/\n  replacement: ./\n""")

    try:
        result = args.func(args)
    except util.UserError as e:
        log.error(str(e))
        sys.exit(1)
    finally:
        log.flush()

    if result is None:
        result = 0

    sys.exit(result)



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd
from orca import orca

from activitysim.abm.tables.size_terms import tour_destination_size_terms
from activitysim.core import (
    config,
    expressions,
    inject,
    logit,
    los,
    pipeline,
    simulate,
    tracing,
)
from activitysim.core.interaction_sample import interaction_sample
from activitysim.core.interaction_sample_simulate import interaction_sample_simulate
from activitysim.core.util import reindex

from . import logsums as logsum
from . import trip

logger = logging.getLogger(__name__)
DUMP = False

# temp column names for presampling
DEST_MAZ = "dest_MAZ"
DEST_TAZ = "dest_TAZ"

# likewise a temp, but if already in choosers,
# we assume we can use it opportunistically
ORIG_TAZ = "orig_TAZ"
ORIG_MAZ = "orig_MAZ"
ORIG_TAZ_EXT = "orig_TAZ_ext"


def get_od_id_col(origin_col, destination_col):
    colname = "{0}_{1}".format(origin_col, destination_col)
    return colname


def create_od_id_col(df, origin_col, destination_col):
    return df[origin_col].astype(str) + "_" + df[destination_col].astype(str)


def _get_od_cols_from_od_id(
    df, orig_col_name=None, dest_col_name=None, od_id_col="choice"
):

    df[orig_col_name] = df[od_id_col].str.split("_").str[0].astype(int)
    df[dest_col_name] = df[od_id_col].str.split("_").str[1].astype(int)

    return df


def _create_od_alts_from_dest_size_terms(
    size_terms_df,
    segment_name,
    od_id_col=None,
    origin_id_col="origin",
    dest_id_col="destination",
    origin_filter=None,
    origin_attr_cols=None,
):
    """
    Extend destination size terms to create dataframe representing the
    cartesian product of tour origins and destinations. Actual "Size Terms"
    will still only be associated with the destinations, but individual
    attributes of the origins can be preserved.
    """

    land_use = inject.get_table("land_use").to_frame(columns=origin_attr_cols)

    if origin_filter:
        origins = land_use.query(origin_filter)
    else:
        origins = land_use

    n_repeat = len(origins)
    od_alts = size_terms_df.reindex(size_terms_df.index.repeat(n_repeat))
    od_alts[origin_id_col] = list(origins.index.values) * od_alts.index.nunique()
    od_alts.reset_index(inplace=True)
    if dest_id_col not in od_alts.columns:
        od_alts.rename(columns={land_use.index.name: dest_id_col}, inplace=True)

    if od_id_col is None:
        new_index_name = get_od_id_col(origin_id_col, dest_id_col)
    else:
        new_index_name = od_id_col
    od_alts[new_index_name] = (
        od_alts[origin_id_col].astype(str) + "_" + od_alts[dest_id_col].astype(str)
    )
    od_alts.set_index(new_index_name, inplace=True)

    # manually add origin attributes to output since these can't be generated by
    # the destination-based size term calculator
    if origin_attr_cols:
        land_use.index.name = origin_id_col
        land_use.reset_index(inplace=True)
        od_alts.reset_index(inplace=True)
        od_alts = pd.merge(
            od_alts,
            land_use[origin_attr_cols + [origin_id_col]],
            on=origin_id_col,
            how="left",
        ).set_index(new_index_name)

    return od_alts


def _od_sample(
    spec_segment_name,
    choosers,
    network_los,
    destination_size_terms,
    origin_id_col,
    dest_id_col,
    skims,
    estimator,
    model_settings,
    alt_od_col_name,
    chunk_size,
    chunk_tag,
    trace_label,
):

    model_spec = simulate.spec_for_segment(
        model_settings,
        spec_id="SAMPLE_SPEC",
        segment_name=spec_segment_name,
        estimator=estimator,
    )
    if alt_od_col_name is None:
        alt_col_name = get_od_id_col(origin_id_col, dest_id_col)
    else:
        alt_col_name = alt_od_col_name

    logger.info("running %s with %d tours", trace_label, len(choosers))

    sample_size = model_settings["SAMPLE_SIZE"]
    if config.setting("disable_destination_sampling", False) or (
        estimator and estimator.want_unsampled_alternatives
    ):
        # FIXME interaction_sample will return unsampled complete alternatives
        # with probs and pick_count
        logger.info(
            (
                "Estimation mode for %s using unsampled alternatives "
                "short_circuit_choices"
            )
            % trace_label
        )
        sample_size = 0

    locals_d = {"skims": skims}
    constants = config.get_model_constants(model_settings)
    if constants is not None:
        locals_d.update(constants)

    origin_filter = model_settings.get("ORIG_FILTER", None)
    origin_attr_cols = model_settings["ORIGIN_ATTR_COLS_TO_USE"]

    od_alts_df = _create_od_alts_from_dest_size_terms(
        destination_size_terms,
        spec_segment_name,
        od_id_col=alt_col_name,
        origin_id_col=origin_id_col,
        dest_id_col=dest_id_col,
        origin_filter=origin_filter,
        origin_attr_cols=origin_attr_cols,
    )

    if skims.orig_key == ORIG_TAZ:
        od_alts_df[ORIG_TAZ] = map_maz_to_taz(od_alts_df[origin_id_col], network_los)

    elif skims.orig_key not in od_alts_df:
        logger.error("Alts df is missing origin skim key column.")

    choices = interaction_sample(
        choosers,
        alternatives=od_alts_df,
        sample_size=sample_size,
        alt_col_name=alt_col_name,
        spec=model_spec,
        skims=skims,
        locals_d=locals_d,
        chunk_size=chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
    )

    return choices


def od_sample(
    spec_segment_name,
    choosers,
    model_settings,
    network_los,
    destination_size_terms,
    estimator,
    chunk_size,
    trace_label,
):

    chunk_tag = "tour_od.sample"

    origin_col_name = model_settings["ORIG_COL_NAME"]
    dest_col_name = model_settings["DEST_COL_NAME"]
    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]

    skim_dict = network_los.get_default_skim_dict()
    skims = skim_dict.wrap(origin_col_name, dest_col_name)

    # the name of the od column to be returned in choices
    alt_od_col_name = get_od_id_col(origin_col_name, dest_col_name)
    choices = _od_sample(
        spec_segment_name,
        choosers,
        network_los,
        destination_size_terms,
        origin_col_name,
        dest_col_name,
        skims,
        estimator,
        model_settings,
        alt_od_col_name,
        chunk_size,
        chunk_tag,
        trace_label,
    )

    choices[origin_col_name] = (
        choices[alt_od_col_name].str.split("_").str[0].astype(int)
    )
    choices[dest_col_name] = choices[alt_od_col_name].str.split("_").str[1].astype(int)

    return choices


def map_maz_to_taz(s, network_los):
    maz_to_taz = network_los.maz_taz_df[["MAZ", "TAZ"]].set_index("MAZ").TAZ
    return s.map(maz_to_taz)


def map_maz_to_ext_taz(s):
    land_use = (
        inject.get_table("land_use").to_frame(columns=["external_TAZ"]).external_TAZ
    )
    return s.map(land_use).astype(int)


def map_maz_to_ext_maz(s):
    land_use = (
        inject.get_table("land_use").to_frame(columns=["external_MAZ"]).external_MAZ
    )
    return s.map(land_use).astype(int)


def map_ext_maz_to_maz(s):
    land_use = (
        inject.get_table("land_use").to_frame(columns=["original_MAZ"]).original_MAZ
    )
    return s.map(land_use).astype(int)


def aggregate_size_terms(dest_size_terms, network_los):

    # aggregate MAZ_size_terms to TAZ_size_terms
    MAZ_size_terms = dest_size_terms.copy()

    # add crosswalk DEST_TAZ column to MAZ_size_terms
    MAZ_size_terms[DEST_TAZ] = map_maz_to_taz(MAZ_size_terms.index, network_los)

    # aggregate to TAZ
    TAZ_size_terms = MAZ_size_terms.groupby(DEST_TAZ).agg({"size_term": "sum"})
    assert not TAZ_size_terms["size_term"].isna().any()

    #           size_term
    # dest_TAZ
    # 2              45.0
    # 3              44.0
    # 4              59.0

    # add crosswalk DEST_TAZ column to MAZ_size_terms
    # MAZ_size_terms = MAZ_size_terms.sort_values([DEST_TAZ, 'size_term'])  # maybe helpful for debugging
    MAZ_size_terms = MAZ_size_terms[[DEST_TAZ, "size_term"]].reset_index(drop=False)
    MAZ_size_terms = MAZ_size_terms.sort_values([DEST_TAZ, "zone_id"]).reset_index(
        drop=True
    )

    #       zone_id  dest_TAZ  size_term
    # 0        6097         2       10.0
    # 1       16421         2       13.0
    # 2       24251         3       14.0

    # print(f"TAZ_size_terms ({TAZ_size_terms.shape})\n{TAZ_size_terms}")
    # print(f"MAZ_size_terms ({MAZ_size_terms.shape})\n{MAZ_size_terms}")

    return MAZ_size_terms, TAZ_size_terms


def choose_MAZ_for_TAZ(
    taz_sample,
    MAZ_size_terms,
    trace_label,
    addtl_col_for_unique_key=None,
    dest_maz_id_col=DEST_MAZ,
):
    """
    Convert taz_sample table with TAZ zone sample choices to a table with a MAZ zone chosen for each TAZ
    choose MAZ probabilistically (proportionally by size_term) from set of MAZ zones in parent TAZ

    Parameters
    ----------
    taz_sample: dataframe with duplicated index <chooser_id_col> and columns: <DEST_TAZ>, prob, pick_count
    MAZ_size_terms: dataframe with unique index and columns: <dest_maz_id_col>, dest_TAZ, size_term
    trace_label: str
    addtl_col_for_unique_key: str of col name to use in addition to destination zone if destination
        zone alone will not be unique per chooser as is the case for joint simulation of tour ODs.

    Returns
    -------
    dataframe with with duplicated index <chooser_id_col> and columns: <DEST_MAZ>, prob, pick_count
    """

    # print(f"taz_sample\n{taz_sample}")
    #           dest_TAZ      prob  pick_count  person_id
    # tour_id
    # 542963          18  0.004778           1      13243
    # 542963          53  0.004224           2      13243
    # 542963          59  0.008628           1      13243

    trace_hh_id = inject.get_injectable("trace_hh_id", None)
    have_trace_targets = trace_hh_id and tracing.has_trace_targets(taz_sample)
    if have_trace_targets:
        trace_label = tracing.extend_trace_label(trace_label, "choose_MAZ_for_TAZ")

        CHOOSER_ID = (
            taz_sample.index.name
        )  # zone_id for tours, but person_id for location choice
        assert CHOOSER_ID is not None

        # write taz choices, pick_counts, probs
        trace_targets = tracing.trace_targets(taz_sample)
        tracing.trace_df(
            taz_sample[trace_targets],
            label=tracing.extend_trace_label(trace_label, "taz_sample"),
            transpose=False,
        )

    if addtl_col_for_unique_key is None:
        addtl_col_for_unique_key = []
    else:
        addtl_col_for_unique_key = [addtl_col_for_unique_key]

    # redupe taz_sample[[DEST_TAZ, 'prob']] using pick_count to repeat rows
    taz_choices = taz_sample[[DEST_TAZ, "prob"] + addtl_col_for_unique_key].reset_index(
        drop=False
    )
    taz_choices = taz_choices.reindex(
        taz_choices.index.repeat(taz_sample.pick_count)
    ).reset_index(drop=True)
    taz_choices = taz_choices.rename(columns={"prob": "TAZ_prob"})

    # print(f"taz_choices\n{taz_choices}")
    #        tour_id  dest_TAZ  TAZ_prob  <addtl_col_for_unique_key>
    # 0         856      4522  0.000679      7066
    # 1         856      4666  0.001222      7066
    # 2         856      4802  0.003473      7066
    # 3         856      4927  0.027282      7066
    # 4         856      4961  0.004853      7066

    # print(f"MAZ_size_terms\n{MAZ_size_terms}")
    #       zone_id  dest_TAZ  size_term
    # 0        6097         2      7.420
    # 1       16421         2      9.646
    # 2       24251         2     10.904

    # just to make it clear we are siloing choices by chooser_id
    chooser_id_col = (
        taz_sample.index.name
    )  # should be canonical chooser index name (e.g. 'person_id')

    # for random_for_df, we need df with de-duplicated chooser canonical index
    chooser_df = pd.DataFrame(index=taz_sample.index[~taz_sample.index.duplicated()])
    num_choosers = len(chooser_df)
    assert chooser_df.index.name == chooser_id_col

    # to make choices, <taz_sample_size> rands for each chooser (one rand for each sampled TAZ)
    # taz_sample_size will be model_settings['SAMPLE_SIZE'] samples, except if we are estimating
    taz_sample_size = taz_choices.groupby(chooser_id_col)[DEST_TAZ].count().max()

    # taz_choices index values should be contiguous
    assert (
        taz_choices[chooser_id_col] == np.repeat(chooser_df.index, taz_sample_size)
    ).all()

    # we need to choose a MAZ for each DEST_TAZ choice
    # probability of choosing MAZ based on MAZ size_term fraction of TAZ total
    # there will be a different set (and number) of candidate MAZs for each TAZ
    # (preserve index, which will have duplicates as result of join)
    # maz_sizes.index is the integer offset into taz_choices of the taz for which the maz_size row is a candidate)
    maz_sizes = pd.merge(
        taz_choices[
            [chooser_id_col, DEST_TAZ] + addtl_col_for_unique_key
        ].reset_index(),
        MAZ_size_terms,
        how="left",
        on=DEST_TAZ,
    ).set_index("index")

    #         tour_id  dest_TAZ  <addtl_col_for_unique_key>  zone_id  size_term
    # index
    # 0          856      4522          7066                  1624    899.648
    # 0          856      4522          7066                  1627     52.918
    # 1          856      4666          7066                  8507    253.256
    # 2          856      4802          7066                  8444    904.930
    # 3          856      4927          7066                  6999    712.630

    if have_trace_targets:
        # write maz_sizes: maz_sizes[index,tour_id,dest_TAZ,zone_id,size_term]

        maz_sizes_trace_targets = tracing.trace_targets(maz_sizes, slicer=CHOOSER_ID)
        trace_maz_sizes = maz_sizes[maz_sizes_trace_targets]
        tracing.trace_df(
            trace_maz_sizes,
            label=tracing.extend_trace_label(trace_label, "maz_sizes"),
            transpose=False,
        )

    # number of DEST_TAZ candidates per chooser
    maz_counts = maz_sizes.groupby(maz_sizes.index).size().values

    # max number of MAZs for any TAZ
    max_maz_count = maz_counts.max()

    # offsets of the first and last rows of each chooser in sparse interaction_utilities
    last_row_offsets = maz_counts.cumsum()
    first_row_offsets = np.insert(last_row_offsets[:-1], 0, 0)

    # repeat the row offsets once for each dummy utility to insert
    # (we want to insert dummy utilities at the END of the list of alternative utilities)
    # inserts is a list of the indices at which we want to do the insertions
    inserts = np.repeat(last_row_offsets, max_maz_count - maz_counts)

    # insert zero filler to pad each alternative set to same size
    padded_maz_sizes = np.insert(maz_sizes.size_term.values, inserts, 0.0).reshape(
        -1, max_maz_count
    )

    # prob array with one row TAZ_choice, one column per alternative
    row_sums = padded_maz_sizes.sum(axis=1)
    maz_probs = np.divide(padded_maz_sizes, row_sums.reshape(-1, 1))
    assert maz_probs.shape == (num_choosers * taz_sample_size, max_maz_count)

    rands = pipeline.get_rn_generator().random_for_df(chooser_df, n=taz_sample_size)
    rands = rands.reshape(-1, 1)
    assert len(rands) == num_choosers * taz_sample_size
    assert len(rands) == maz_probs.shape[0]

    # make choices
    # positions is array with the chosen alternative represented as a column index in probs
    # which is an integer between zero and max_maz_count
    positions = np.argmax((maz_probs.cumsum(axis=1) - rands) > 0.0, axis=1)

    # shouldn't have chosen any of the dummy pad positions
    assert (positions < maz_counts).all()

    # this take can choose same dest more than once (choice with replacement)
    # this is why you need the groupby later on, which is why the prob is a max
    taz_choices[DEST_MAZ] = maz_sizes["zone_id"].take(positions + first_row_offsets)
    taz_choices["MAZ_prob"] = maz_probs[np.arange(maz_probs.shape[0]), positions]
    taz_choices["prob"] = taz_choices["TAZ_prob"] * taz_choices["MAZ_prob"]

    if have_trace_targets:

        taz_choices_trace_targets = tracing.trace_targets(
            taz_choices, slicer=CHOOSER_ID
        )
        trace_taz_choices_df = taz_choices[taz_choices_trace_targets]
        tracing.trace_df(
            trace_taz_choices_df,
            label=tracing.extend_trace_label(trace_label, "taz_choices"),
            transpose=False,
        )

        lhs_df = trace_taz_choices_df[[CHOOSER_ID, DEST_TAZ] + addtl_col_for_unique_key]
        alt_dest_columns = [f"dest_maz_{c}" for c in range(max_maz_count)]

        # following the same logic as the full code, but for trace cutout
        trace_maz_counts = maz_counts[taz_choices_trace_targets]
        trace_last_row_offsets = maz_counts[taz_choices_trace_targets].cumsum()
        trace_inserts = np.repeat(
            trace_last_row_offsets, max_maz_count - trace_maz_counts
        )

        # trace dest_maz_alts
        padded_maz_sizes = np.insert(
            trace_maz_sizes[CHOOSER_ID].values, trace_inserts, 0.0
        ).reshape(-1, max_maz_count)
        df = pd.DataFrame(
            data=padded_maz_sizes,
            columns=alt_dest_columns,
            index=trace_taz_choices_df.index,
        )
        df = pd.concat([lhs_df, df], axis=1)
        tracing.trace_df(
            df,
            label=tracing.extend_trace_label(trace_label, "dest_maz_alts"),
            transpose=False,
        )

        # trace dest_maz_size_terms
        padded_maz_sizes = np.insert(
            trace_maz_sizes["size_term"].values, trace_inserts, 0.0
        ).reshape(-1, max_maz_count)
        df = pd.DataFrame(
            data=padded_maz_sizes,
            columns=alt_dest_columns,
            index=trace_taz_choices_df.index,
        )
        df = pd.concat([lhs_df, df], axis=1)
        tracing.trace_df(
            df,
            label=tracing.extend_trace_label(trace_label, "dest_maz_size_terms"),
            transpose=False,
        )

        # trace dest_maz_probs
        df = pd.DataFrame(
            data=maz_probs[taz_choices_trace_targets],
            columns=alt_dest_columns,
            index=trace_taz_choices_df.index,
        )
        df = pd.concat([lhs_df, df], axis=1)
        df["rand"] = rands[taz_choices_trace_targets]
        tracing.trace_df(
            df,
            label=tracing.extend_trace_label(trace_label, "dest_maz_probs"),
            transpose=False,
        )

    taz_choices = taz_choices.drop(columns=["TAZ_prob", "MAZ_prob"])
    taz_choices_w_maz = taz_choices.groupby(
        [chooser_id_col, dest_maz_id_col] + addtl_col_for_unique_key
    ).agg(prob=("prob", "first"), pick_count=("prob", "count"))

    taz_choices_w_maz.reset_index(inplace=True)
    taz_choices_w_maz.set_index(chooser_id_col, inplace=True)

    return taz_choices_w_maz


def od_presample(
    spec_segment_name,
    choosers,
    model_settings,
    network_los,
    destination_size_terms,
    estimator,
    chunk_size,
    trace_label,
):

    trace_label = tracing.extend_trace_label(trace_label, "presample")
    chunk_tag = "tour_od.presample"

    logger.info(f"{trace_label} od_presample")

    alt_od_col_name = get_od_id_col(ORIG_MAZ, DEST_TAZ)

    MAZ_size_terms, TAZ_size_terms = aggregate_size_terms(
        destination_size_terms, network_los
    )

    # create wrapper with keys for this lookup - in this case there is a ORIG_TAZ
    # in the choosers and a DEST_TAZ in the alternatives which get merged during
    # interaction the skims will be available under the name "skims" for any @ expressions
    skim_dict = network_los.get_skim_dict("taz")
    skims = skim_dict.wrap(ORIG_TAZ, DEST_TAZ)

    orig_MAZ_dest_TAZ_sample = _od_sample(
        spec_segment_name,
        choosers,
        network_los,
        TAZ_size_terms,
        ORIG_MAZ,
        DEST_TAZ,
        skims,
        estimator,
        model_settings,
        alt_od_col_name,
        chunk_size,
        chunk_tag,
        trace_label,
    )

    orig_MAZ_dest_TAZ_sample[ORIG_MAZ] = (
        orig_MAZ_dest_TAZ_sample[alt_od_col_name].str.split("_").str[0].astype(int)
    )
    orig_MAZ_dest_TAZ_sample[DEST_TAZ] = (
        orig_MAZ_dest_TAZ_sample[alt_od_col_name].str.split("_").str[1].astype(int)
    )

    # choose a MAZ for each DEST_TAZ choice, choice probability based on
    # MAZ size_term fraction of TAZ total

    maz_choices = choose_MAZ_for_TAZ(
        orig_MAZ_dest_TAZ_sample,
        MAZ_size_terms,
        trace_label,
        addtl_col_for_unique_key=ORIG_MAZ,
    )

    # outputs
    assert DEST_MAZ in maz_choices

    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]
    chooser_orig_col_name = model_settings["CHOOSER_ORIG_COL_NAME"]
    maz_choices = maz_choices.rename(
        columns={DEST_MAZ: alt_dest_col_name, ORIG_MAZ: chooser_orig_col_name}
    )

    return maz_choices


class SizeTermCalculator(object):
    """
    convenience object to provide size_terms for a selector (e.g.
    non_mandatory) for various segments (e.g. tour_type or purpose)
    returns size terms for specified segment in df or series form.
    """

    def __init__(self, size_term_selector):

        # do this once so they can request size_terms for various segments (tour_type or purpose)
        land_use = inject.get_table("land_use")
        self.land_use = land_use
        size_terms = inject.get_injectable("size_terms")
        self.destination_size_terms = tour_destination_size_terms(
            self.land_use, size_terms, size_term_selector
        )

        assert not self.destination_size_terms.isna().any(axis=None)

    def omnibus_size_terms_df(self):
        return self.destination_size_terms

    def dest_size_terms_df(self, segment_name, trace_label):
        # return size terms as df with one column named 'size_term'
        # convenient if creating or merging with alts

        size_terms = self.destination_size_terms[[segment_name]].copy()
        size_terms.columns = ["size_term"]

        # FIXME - no point in considering impossible alternatives (where dest size term is zero)
        logger.debug(
            f"SizeTermCalculator dropping {(~(size_terms.size_term > 0)).sum()} "
            f"of {len(size_terms)} rows where size_term is zero for {segment_name}"
        )
        size_terms = size_terms[size_terms.size_term > 0]

        if len(size_terms) == 0:
            logger.warning(
                f"SizeTermCalculator: no zones with non-zero size terms for {segment_name} in {trace_label}"
            )

        return size_terms

    def dest_size_terms_series(self, segment_name):
        # return size terms as as series
        # convenient (and no copy overhead) if reindexing and assigning into alts column
        return self.destination_size_terms[segment_name]


def run_od_sample(
    spec_segment_name,
    tours,
    model_settings,
    network_los,
    destination_size_terms,
    estimator,
    chunk_size,
    trace_label,
):

    model_spec = simulate.spec_for_segment(
        model_settings,
        spec_id="SAMPLE_SPEC",
        segment_name=spec_segment_name,
        estimator=estimator,
    )

    choosers = tours
    # FIXME - MEMORY HACK - only include columns actually used in spec
    chooser_columns = model_settings["SIMULATE_CHOOSER_COLUMNS"]
    choosers = choosers[chooser_columns]

    # interaction_sample requires that choosers.index.is_monotonic_increasing
    if not choosers.index.is_monotonic_increasing:
        logger.debug(
            f"run_destination_sample {trace_label} sorting choosers because not monotonic_increasing"
        )
        choosers = choosers.sort_index()

    # by default, enable presampling for multizone systems, unless they disable it in settings file
    pre_sample_taz = not (network_los.zone_system == los.ONE_ZONE)
    if pre_sample_taz and not config.setting("want_dest_choice_presampling", True):
        pre_sample_taz = False
        logger.info(
            f"Disabled destination zone presampling for {trace_label} "
            f"because 'want_dest_choice_presampling' setting is False"
        )

    if pre_sample_taz:

        logger.info(
            "Running %s destination_presample with %d tours" % (trace_label, len(tours))
        )

        choices = od_presample(
            spec_segment_name,
            choosers,
            model_settings,
            network_los,
            destination_size_terms,
            estimator,
            chunk_size,
            trace_label,
        )

    else:
        choices = od_sample(
            spec_segment_name,
            choosers,
            model_settings,
            network_los,
            destination_size_terms,
            estimator,
            chunk_size,
            trace_label,
        )

    return choices


def run_od_logsums(
    spec_segment_name,
    tours_merged_df,
    od_sample,
    model_settings,
    network_los,
    estimator,
    chunk_size,
    trace_hh_id,
    trace_label,
):
    """
    add logsum column to existing tour_destination_sample table

    logsum is calculated by running the mode_choice model for each sample
    (person, OD_id) pair in od_sample, and computing the logsum of all the utilities
    """
    chunk_tag = "tour_od.logsums"
    logsum_settings = config.read_model_settings(model_settings["LOGSUM_SETTINGS"])
    origin_id_col = model_settings["ORIG_COL_NAME"]
    dest_id_col = model_settings["DEST_COL_NAME"]
    tour_od_id_col = get_od_id_col(origin_id_col, dest_id_col)

    # FIXME - MEMORY HACK - only include columns actually used in spec
    tours_merged_df = logsum.filter_chooser_columns(
        tours_merged_df, logsum_settings, model_settings
    )

    # merge ods into choosers table
    choosers = od_sample.join(tours_merged_df, how="left")
    choosers[tour_od_id_col] = (
        choosers[origin_id_col].astype(str) + "_" + choosers[dest_id_col].astype(str)
    )

    logger.info("Running %s with %s rows", trace_label, len(choosers))

    tracing.dump_df(DUMP, choosers, trace_label, "choosers")

    # run trip mode choice to compute tour mode choice logsums
    if logsum_settings.get("COMPUTE_TRIP_MODE_CHOICE_LOGSUMS", False):

        pseudo_tours = choosers.copy()
        trip_mode_choice_settings = config.read_model_settings("trip_mode_choice")

        # tours_merged table doesn't yet have all the cols it needs to be called (e.g.
        # home_zone_id), so in order to compute tour mode choice/trip mode choice logsums
        # in this step we have to pass all tour-level attributes in with the main trips
        # table. see trip_mode_choice.py L56-61 for more details.
        tour_cols_needed = trip_mode_choice_settings.get(
            "TOURS_MERGED_CHOOSER_COLUMNS", []
        )
        tour_cols_needed.append(tour_od_id_col)

        # from tour_mode_choice.py
        not_university = (
            pseudo_tours.tour_type != "school"
        ) | ~pseudo_tours.is_university
        pseudo_tours["tour_purpose"] = pseudo_tours.tour_type.where(
            not_university, "univ"
        )

        pseudo_tours["stop_frequency"] = "0out_0in"
        pseudo_tours["primary_purpose"] = pseudo_tours["tour_purpose"]
        choosers_og_index = choosers.index.name
        pseudo_tours.reset_index(inplace=True)
        pseudo_tours.index.name = "unique_id"

        # need dest_id_col to create dest col in trips, but need to preserve
        # tour dest as separate column in the trips table bc the trip mode choice
        # preprocessor isn't able to get the tour dest from the tours table bc the
        # tours don't yet have ODs.
        stop_frequency_alts = inject.get_injectable("stop_frequency_alts")
        pseudo_tours["tour_destination"] = pseudo_tours[dest_id_col]
        trips = trip.initialize_from_tours(
            pseudo_tours,
            stop_frequency_alts,
            [origin_id_col, dest_id_col, "tour_destination", "unique_id"],
        )
        outbound = trips["outbound"]
        trips["depart"] = reindex(pseudo_tours.start, trips.unique_id)
        trips.loc[~outbound, "depart"] = reindex(
            pseudo_tours.end, trips.loc[~outbound, "unique_id"]
        )

        logsum_trips = pd.DataFrame()
        nest_spec = config.get_logit_model_settings(logsum_settings)

        # actual coeffs dont matter here, just need them to load the nest structure
        coefficients = simulate.get_segment_coefficients(
            logsum_settings, pseudo_tours.iloc[0]["tour_purpose"]
        )
        nest_spec = simulate.eval_nest_coefficients(
            nest_spec, coefficients, trace_label
        )
        tour_mode_alts = []
        for nest in logit.each_nest(nest_spec):
            if nest.is_leaf:
                tour_mode_alts.append(nest.name)

        # repeat rows from the trips table iterating over tour mode
        for tour_mode in tour_mode_alts:
            trips["tour_mode"] = tour_mode
            logsum_trips = pd.concat((logsum_trips, trips), ignore_index=True)
        assert len(logsum_trips) == len(trips) * len(tour_mode_alts)
        logsum_trips.index.name = "trip_id"

        for col in tour_cols_needed:
            if col not in trips:
                logsum_trips[col] = reindex(pseudo_tours[col], logsum_trips.unique_id)

        pipeline.replace_table("trips", logsum_trips)
        tracing.register_traceable_table("trips", logsum_trips)
        pipeline.get_rn_generator().add_channel("trips", logsum_trips)

        # run trip mode choice on pseudo-trips. use orca instead of pipeline to
        # execute the step because pipeline can only handle one open step at a time
        orca.run(["trip_mode_choice"])

        # grab trip mode choice logsums and pivot by tour mode and direction, index
        # on tour_id to enable merge back to choosers table
        trips = inject.get_table("trips").to_frame()
        trip_dir_mode_logsums = trips.pivot(
            index=["tour_id", tour_od_id_col],
            columns=["tour_mode", "outbound"],
            values="trip_mode_choice_logsum",
        )
        new_cols = [
            "_".join(["logsum", mode, "outbound" if outbound else "inbound"])
            for mode, outbound in trip_dir_mode_logsums.columns
        ]
        trip_dir_mode_logsums.columns = new_cols

        choosers.reset_index(inplace=True)
        choosers.set_index(["tour_id", tour_od_id_col], inplace=True)
        choosers = pd.merge(
            choosers, trip_dir_mode_logsums, left_index=True, right_index=True
        )
        choosers.reset_index(inplace=True)
        choosers.set_index(choosers_og_index, inplace=True)

        pipeline.get_rn_generator().drop_channel("trips")
        tracing.deregister_traceable_table("trips")

        assert (od_sample.index == choosers.index).all()
        for col in new_cols:
            od_sample[col] = choosers[col]

    logsums = logsum.compute_logsums(
        choosers,
        spec_segment_name,
        logsum_settings,
        model_settings,
        network_los,
        chunk_size,
        chunk_tag,
        trace_label,
        "end",
        "start",
        "duration",
    )

    assert (od_sample.index == logsums.index).all()
    od_sample["tour_mode_choice_logsum"] = logsums

    return od_sample


def run_od_simulate(
    spec_segment_name,
    tours,
    od_sample,
    want_logsums,
    model_settings,
    network_los,
    destination_size_terms,
    estimator,
    chunk_size,
    trace_label,
):
    """
    run simulate OD choices on tour_od_sample annotated with mode_choice
    logsum to select a tour OD from sample alternatives
    """

    model_spec = simulate.spec_for_segment(
        model_settings,
        spec_id="SPEC",
        segment_name=spec_segment_name,
        estimator=estimator,
    )

    # merge persons into tours
    choosers = tours

    # FIXME - MEMORY HACK - only include columns actually used in spec
    chooser_columns = model_settings["SIMULATE_CHOOSER_COLUMNS"]
    choosers = choosers[chooser_columns]

    # interaction_sample requires that choosers.index.is_monotonic_increasing
    if not choosers.index.is_monotonic_increasing:
        logger.debug(
            f"run_destination_simulate {trace_label} sorting choosers because not monotonic_increasing"
        )
        choosers = choosers.sort_index()

    if estimator:
        estimator.write_choosers(choosers)

    origin_col_name = model_settings["ORIG_COL_NAME"]
    dest_col_name = model_settings["DEST_COL_NAME"]
    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]
    origin_attr_cols = model_settings["ORIGIN_ATTR_COLS_TO_USE"]

    alt_od_col_name = get_od_id_col(origin_col_name, dest_col_name)
    od_sample[alt_od_col_name] = create_od_id_col(
        od_sample, origin_col_name, dest_col_name
    )

    # alternatives are pre-sampled and annotated with logsums and pick_count
    # but we have to merge size_terms column into alt sample list
    od_sample["size_term"] = reindex(
        destination_size_terms.size_term, od_sample[alt_dest_col_name]
    )

    # also have to add origin attribute columns
    lu = inject.get_table("land_use").to_frame(columns=origin_attr_cols)
    od_sample = pd.merge(
        od_sample, lu, left_on=origin_col_name, right_index=True, how="left"
    )

    tracing.dump_df(DUMP, od_sample, trace_label, "alternatives")

    constants = config.get_model_constants(model_settings)

    logger.info("Running tour_destination_simulate with %d persons", len(choosers))

    # create wrapper with keys for this lookup - in this case there is an origin ID
    # column and a destination ID columns in the alternatives table.
    # the skims will be available under the name "skims" for any @ expressions
    skim_dict = network_los.get_default_skim_dict()
    skims = skim_dict.wrap(origin_col_name, dest_col_name)

    locals_d = {
        "skims": skims,
    }
    if constants is not None:
        locals_d.update(constants)

    tracing.dump_df(DUMP, choosers, trace_label, "choosers")
    choices = interaction_sample_simulate(
        choosers,
        od_sample,
        spec=model_spec,
        choice_column=alt_od_col_name,
        want_logsums=want_logsums,
        skims=skims,
        locals_d=locals_d,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name="origin_destination",
        estimator=estimator,
    )

    if not want_logsums:
        choices = choices.to_frame("choice")

    choices = _get_od_cols_from_od_id(choices, origin_col_name, dest_col_name)

    return choices


def run_tour_od(
    tours,
    persons,
    want_logsums,
    want_sample_table,
    model_settings,
    network_los,
    estimator,
    chunk_size,
    trace_hh_id,
    trace_label,
):

    size_term_calculator = SizeTermCalculator(model_settings["SIZE_TERM_SELECTOR"])
    preprocessor_settings = model_settings.get("preprocessor", None)
    origin_col_name = model_settings["ORIG_COL_NAME"]

    chooser_segment_column = model_settings["CHOOSER_SEGMENT_COLUMN_NAME"]

    # maps segment names to compact (integer) ids
    segments = model_settings["SEGMENTS"]

    # interaction_sample_simulate insists choosers appear in same order as alts
    tours = tours.sort_index()

    choices_list = []
    sample_list = []
    for segment_name in segments:

        choosers = tours[tours[chooser_segment_column] == segment_name]

        choosers = pd.merge(
            choosers,
            persons.to_frame(columns=["is_university", "demographic_segment"]),
            left_on="person_id",
            right_index=True,
        )

        # - annotate choosers
        if preprocessor_settings:
            expressions.assign_columns(
                df=choosers,
                model_settings=preprocessor_settings,
                trace_label=trace_label,
            )

        # size_term segment is segment_name
        segment_destination_size_terms = size_term_calculator.dest_size_terms_df(
            segment_name, trace_label
        )

        if choosers.shape[0] == 0:
            logger.info(
                "%s skipping segment %s: no choosers", trace_label, segment_name
            )
            continue

        # - od_sample
        spec_segment_name = segment_name  # spec_segment_name is segment_name

        od_sample_df = run_od_sample(
            spec_segment_name,
            choosers,
            model_settings,
            network_los,
            segment_destination_size_terms,
            estimator,
            chunk_size=chunk_size,
            trace_label=tracing.extend_trace_label(
                trace_label, "sample.%s" % segment_name
            ),
        )

        if model_settings["ORIG_FILTER"] == "original_MAZ > 0":
            pass
        elif model_settings["ORIG_FILTER"] == "external_TAZ > 0":
            # sampled alts using internal mazs, so now we
            # have to convert to using the external tazs
            od_sample_df[origin_col_name] = map_maz_to_ext_maz(
                od_sample_df[origin_col_name]
            )
        else:
            raise ValueError(
                "Not sure how you identified tour origins but you probably need "
                "to choose a different ORIG_FILTER setting"
            )

        # - destination_logsums
        od_sample_df = run_od_logsums(
            spec_segment_name,
            choosers,
            od_sample_df,
            model_settings,
            network_los,
            estimator,
            chunk_size=chunk_size,
            trace_hh_id=trace_hh_id,
            trace_label=tracing.extend_trace_label(
                trace_label, "logsums.%s" % segment_name
            ),
        )

        # - od_simulate
        choices = run_od_simulate(
            spec_segment_name,
            choosers,
            od_sample_df,
            want_logsums=want_logsums,
            model_settings=model_settings,
            network_los=network_los,
            destination_size_terms=segment_destination_size_terms,
            estimator=estimator,
            chunk_size=chunk_size,
            trace_label=tracing.extend_trace_label(
                trace_label, "simulate.%s" % segment_name
            ),
        )

        choices_list.append(choices)
        if estimator:
            assert estimator.want_unsampled_alternatives

        if want_sample_table:
            # FIXME - sample_table
            od_sample_df.set_index(
                model_settings["ALT_DEST_COL_NAME"], append=True, inplace=True
            )
            sample_list.append(od_sample_df)

        # FIXME - want to do this here?
        del od_sample_df

    if len(choices_list) > 0:
        choices_df = pd.concat(choices_list)

    if len(sample_list) > 0:
        save_sample_df = pd.concat(sample_list)
    else:
        # this could happen either with small samples as above, or if no saved sample desired
        save_sample_df = None

    return choices_df, save_sample_df



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import config, expressions, inject, pipeline, simulate, tracing

from .util import estimation
from .util.overlap import hh_time_window_overlap

logger = logging.getLogger(__name__)


def add_null_results(trace_label, tours):
    logger.info("Skipping %s: add_null_results" % trace_label)
    tours["composition"] = ""
    pipeline.replace_table("tours", tours)


@inject.step()
def joint_tour_composition(tours, households, persons, chunk_size, trace_hh_id):
    """
    This model predicts the makeup of the travel party (adults, children, or mixed).
    """
    trace_label = "joint_tour_composition"
    model_settings_file_name = "joint_tour_composition.yaml"

    tours = tours.to_frame()
    joint_tours = tours[tours.tour_category == "joint"]

    # - if no joint tours
    if joint_tours.shape[0] == 0:
        add_null_results(trace_label, tours)
        return

    model_settings = config.read_model_settings(model_settings_file_name)
    estimator = estimation.manager.begin_estimation("joint_tour_composition")

    # - only interested in households with joint_tours
    households = households.to_frame()
    households = households[households.num_hh_joint_tours > 0]

    persons = persons.to_frame()
    persons = persons[persons.household_id.isin(households.index)]

    logger.info(
        "Running joint_tour_composition with %d joint tours" % joint_tours.shape[0]
    )

    # - run preprocessor
    preprocessor_settings = model_settings.get("preprocessor", None)
    if preprocessor_settings:

        locals_dict = {
            "persons": persons,
            "hh_time_window_overlap": hh_time_window_overlap,
        }

        expressions.assign_columns(
            df=households,
            model_settings=preprocessor_settings,
            locals_dict=locals_dict,
            trace_label=trace_label,
        )

    joint_tours_merged = pd.merge(
        joint_tours, households, left_on="household_id", right_index=True, how="left"
    )

    # - simple_simulate
    model_spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

    nest_spec = config.get_logit_model_settings(model_settings)
    constants = config.get_model_constants(model_settings)

    if estimator:
        estimator.write_spec(model_settings)
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_coefficients(coefficients_df, model_settings)
        estimator.write_choosers(joint_tours_merged)

    choices = simulate.simple_simulate(
        choosers=joint_tours_merged,
        spec=model_spec,
        nest_spec=nest_spec,
        locals_d=constants,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name="composition",
        estimator=estimator,
    )

    # convert indexes to alternative names
    choices = pd.Series(model_spec.columns[choices.values], index=choices.index)

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(choices, "tours", "composition")
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    # add composition column to tours for tracing
    joint_tours["composition"] = choices

    # reindex since we ran model on a subset of households
    tours["composition"] = choices.reindex(tours.index).fillna("").astype(str)
    pipeline.replace_table("tours", tours)

    tracing.print_summary(
        "joint_tour_composition", joint_tours.composition, value_counts=True
    )

    if trace_hh_id:
        tracing.trace_df(
            joint_tours,
            label="joint_tour_composition.joint_tours",
            slicer="household_id",
        )



import glob
import logging
import logging.handlers
import os
import traceback

import numpy as np
import pandas as pd
import yaml

from ..cli.create import get_example
from ..cli.run import INJECTABLES, config, pipeline
from ..core import inject, tracing
from ..core.pipeline import open_pipeline, run_model
from . import workspace

logger = logging.getLogger(__name__)


def reload_settings(settings_filename, **kwargs):
    settings = config.read_settings_file(settings_filename, mandatory=True)
    for k in kwargs:
        settings[k] = kwargs[k]
    inject.add_injectable("settings", settings)
    return settings


def component_logging(component_name):
    root_logger = logging.getLogger()

    CLOG_FMT = "%(asctime)s %(levelname)7s - %(name)s: %(message)s"

    logfilename = config.log_file_path(f"asv-{component_name}.log")

    # avoid creation of multiple file handlers for logging components
    # as we will re-enter this function for every component run
    for entry in root_logger.handlers:
        if (isinstance(entry, logging.handlers.RotatingFileHandler)) and (
            entry.formatter._fmt == CLOG_FMT
        ):
            return

    tracing.config_logger(basic=True)
    file_handler = logging.handlers.RotatingFileHandler(
        filename=logfilename,
        mode="a",
        maxBytes=50_000_000,
        backupCount=5,
    )
    formatter = logging.Formatter(
        fmt=CLOG_FMT,
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    file_handler.setFormatter(formatter)
    logging.getLogger().addHandler(file_handler)


def setup_component(
    component_name,
    working_dir=".",
    preload_injectables=(),
    configs_dirs=("configs"),
    data_dir="data",
    output_dir="output",
    settings_filename="settings.yaml",
):
    """
    Prepare to benchmark a model component.

    This function sets up everything, opens the pipeline, and
    reloads table state from checkpoints of prior components.
    All this happens here, before the model component itself
    is actually executed inside the timed portion of the loop.
    """
    if isinstance(configs_dirs, str):
        configs_dirs = [configs_dirs]
    inject.add_injectable(
        "configs_dir", [os.path.join(working_dir, i) for i in configs_dirs]
    )
    inject.add_injectable("data_dir", os.path.join(working_dir, data_dir))
    inject.add_injectable("output_dir", os.path.join(working_dir, output_dir))

    reload_settings(
        settings_filename,
        benchmarking=component_name,
        checkpoints=False,
    )

    component_logging(component_name)
    logger.info("connected to component logger")
    config.filter_warnings()
    logging.captureWarnings(capture=True)

    # register abm steps and other abm-specific injectables outside of
    # benchmark timing loop
    if not inject.is_injectable("preload_injectables"):
        logger.info("preload_injectables yes import")
        from activitysim import abm
    else:
        logger.info("preload_injectables no import")

    # Extract the resume_after argument based on the model immediately
    # prior to the component being benchmarked.
    models = config.setting("models")
    try:
        component_index = models.index(component_name)
    except ValueError:
        # the last component to be benchmarked isn't included in the
        # pre-checkpointed model list, we just resume from the end
        component_index = len(models)
    if component_index:
        resume_after = models[component_index - 1]
    else:
        resume_after = None

    if config.setting("multiprocess", False):
        raise NotImplementedError(
            "multiprocess component benchmarking is not yet implemented"
        )
        # Component level timings for multiprocess benchmarking
        # are not generated using this code that re-runs individual
        # components.  Instead, those benchmarks are generated in
        # aggregate during setup and then extracted from logs later.
    else:
        open_pipeline(resume_after, mode="r")

    for k in preload_injectables:
        if inject.get_injectable(k, None) is not None:
            logger.info("pre-loaded %s", k)

    # Directories Logging
    for k in ["configs_dir", "settings_file_name", "data_dir", "output_dir"]:
        logger.info(f"DIRECTORY {k}: {inject.get_injectable(k, None)}")

    # Settings Logging
    log_settings = [
        "checkpoints",
        "chunk_training_mode",
        "chunk_size",
        "chunk_method",
        "trace_hh_id",
        "households_sample_size",
        "check_for_variability",
        "use_shadow_pricing",
        "want_dest_choice_sample_tables",
        "log_alt_losers",
    ]
    for k in log_settings:
        logger.info(f"SETTING {k}: {config.setting(k)}")

    logger.info("setup_component completed: %s", component_name)


def run_component(component_name):
    logger.info("run_component: %s", component_name)
    try:
        if config.setting("multiprocess", False):
            raise NotImplementedError(
                "multiprocess component benchmarking is not yet implemented"
            )
            # Component level timings for multiprocess benchmarking
            # are not generated using this code that re-runs individual
            # components.  Instead, those benchmarks are generated in
            # aggregate during setup and then extracted from logs later.
        else:
            run_model(component_name)
    except Exception as err:
        logger.exception("run_component exception: %s", component_name)
        raise
    else:
        logger.info("run_component completed: %s", component_name)
    return 0


def teardown_component(component_name):
    logger.info("teardown_component: %s", component_name)

    # use the pipeline module to clear out all the orca tables, so
    # the next benchmark run has a clean slate.
    # anything needed should be reloaded from the pipeline checkpoint file
    pipeline_tables = pipeline.registered_tables()
    for table_name in pipeline_tables:
        logger.info("dropping table %s", table_name)
        pipeline.drop_table(table_name)

    if config.setting("multiprocess", False):
        raise NotImplementedError("multiprocess benchmarking is not yet implemented")
    else:
        pipeline.close_pipeline()
    logger.critical(
        "teardown_component completed: %s\n\n%s\n\n", component_name, "~" * 88
    )
    return 0


def pre_run(
    model_working_dir,
    configs_dirs=None,
    data_dir="data",
    output_dir="output",
    settings_file_name=None,
):
    """
    Pre-run the models, checkpointing everything.

    By checkpointing everything, it is possible to run each benchmark
    by recreating the state of the pipeline immediately prior to that
    component.

    Parameters
    ----------
    model_working_dir : str
        Path to the model working directory, generally inside the
        benchmarking workspace.
    configs_dirs : Iterable[str], optional
        Override the config dirs, similar to using -c on the command line
        for a model run.
    data_dir : str, optional
        Override the data directory similar to using -d on the command line
        for a model run.
    output_dir : str, optional
        Override the output directory similar to using -o on the command line
        for a model run.
    settings_file_name : str, optional
        Override the settings file name, similar to using -s on the command line
        for a model run.
    """
    if configs_dirs is None:
        inject.add_injectable("configs_dir", os.path.join(model_working_dir, "configs"))
    else:
        configs_dirs_ = [os.path.join(model_working_dir, i) for i in configs_dirs]
        inject.add_injectable("configs_dir", configs_dirs_)
    inject.add_injectable("data_dir", os.path.join(model_working_dir, data_dir))
    inject.add_injectable("output_dir", os.path.join(model_working_dir, output_dir))

    if settings_file_name is not None:
        inject.add_injectable("settings_file_name", settings_file_name)

    # Always pre_run from the beginning
    config.override_setting("resume_after", None)

    # register abm steps and other abm-specific injectables
    if not inject.is_injectable("preload_injectables"):
        from activitysim import (  # register abm steps and other abm-specific injectables
            abm,
        )

    if settings_file_name is not None:
        inject.add_injectable("settings_file_name", settings_file_name)

    # cleanup
    # cleanup_output_files()

    tracing.config_logger(basic=False)
    config.filter_warnings()
    logging.captureWarnings(capture=True)

    # directories
    for k in ["configs_dir", "settings_file_name", "data_dir", "output_dir"]:
        logger.info("SETTING %s: %s" % (k, inject.get_injectable(k, None)))

    log_settings = inject.get_injectable("log_settings", {})
    for k in log_settings:
        logger.info("SETTING %s: %s" % (k, config.setting(k)))

    # OMP_NUM_THREADS: openmp
    # OPENBLAS_NUM_THREADS: openblas
    # MKL_NUM_THREADS: mkl
    for env in ["MKL_NUM_THREADS", "OMP_NUM_THREADS", "OPENBLAS_NUM_THREADS"]:
        logger.info(f"ENV {env}: {os.getenv(env)}")

    np_info_keys = [
        "atlas_blas_info",
        "atlas_blas_threads_info",
        "atlas_info",
        "atlas_threads_info",
        "blas_info",
        "blas_mkl_info",
        "blas_opt_info",
        "lapack_info",
        "lapack_mkl_info",
        "lapack_opt_info",
        "mkl_info",
    ]

    for cfg_key in np_info_keys:
        info = np.__config__.get_info(cfg_key)
        if info:
            for info_key in ["libraries"]:
                if info_key in info:
                    logger.info(f"NUMPY {cfg_key} {info_key}: {info[info_key]}")

    t0 = tracing.print_elapsed_time()

    logger.info(f"MODELS: {config.setting('models')}")

    if config.setting("multiprocess", False):
        logger.info("run multi-process complete simulation")
    else:
        logger.info("run single process simulation")
        pipeline.run(models=config.setting("models"))
        pipeline.close_pipeline()

    tracing.print_elapsed_time("prerun required models for checkpointing", t0)

    return 0


def run_multiprocess():
    logger.info("run multiprocess simulation")
    tracing.delete_trace_files()
    tracing.delete_output_files("h5")
    tracing.delete_output_files("csv")
    tracing.delete_output_files("txt")
    tracing.delete_output_files("yaml")
    tracing.delete_output_files("prof")
    tracing.delete_output_files("omx")

    from activitysim.core import mp_tasks

    injectables = {k: inject.get_injectable(k) for k in INJECTABLES}
    mp_tasks.run_multiprocess(injectables)

    assert not pipeline.is_open()

    if config.setting("cleanup_pipeline_after_run", False):
        pipeline.cleanup_pipeline()


########


def local_dir():
    benchmarking_directory = workspace.get_dir()
    if benchmarking_directory is not None:
        return benchmarking_directory
    return os.getcwd()


def model_dir(*subdirs):
    return os.path.join(local_dir(), "models", *subdirs)


def template_setup_cache(
    example_name,
    component_names,
    benchmark_settings,
    benchmark_network_los,
    config_dirs=("configs",),
    data_dir="data",
    output_dir="output",
    settings_filename="settings.yaml",
    skip_component_names=None,
    config_overload_dir="dynamic_configs",
):
    """
    Prepare an example model for benchmarking.

    The algorithm for benchmarking components requires an existing pipeline
    file with checkpoints after every component, which allows the code to
    recreate the state of pipeline tables immediately prior to each component's
    execution.  The checkpoints file is very large, and can be recreated
    by running the model, so it is not stored/downloaded but just rebuilt as
    needed.

    This template function creates that pipeline store if it does not exist by
    getting the example model and running once through from the beginning through
    the last component to be benchmarked.  After this is done, a token is written
    out to `benchmark-setup-token.txt` in the output directory, flagging that
    the complete checkpointed pipeline has been created without errors and is
    ready to use.  If the template setup finds this token file, all this work
    is assumed to be already done and is skipped.

    Parameters
    ----------
    example_name : str
        The name of the example to benchmark, as used in
        the `activitysim create` command.
    component_names : Sequence
        The names of the model components to be individually
        benchmarked.  This list does not need to include all
        the components usually included in the example.
    benchmark_settings : Mapping
        Settings values to override from their usual values
        in the example.
    benchmark_network_los : Mapping
        Network LOS values to override from their usual values
        in the example.
    config_dirs : Sequence
    data_dir : str
    output_dir : str
    settings_filename : str
    skip_component_names : Sequence, optional
        Skip running these components when setting up the
        benchmarks (i.e. in pre-run).
    config_overload_dir : str, default 'dynamic_configs'
    """
    try:
        os.makedirs(model_dir(), exist_ok=True)
        get_example(
            example_name=example_name,
            destination=model_dir(),
            benchmarking=True,
        )
        os.makedirs(model_dir(example_name, config_overload_dir), exist_ok=True)

        # Find the settings file and extract the complete set of models included
        from ..core.config import read_settings_file

        try:
            existing_settings, settings_filenames = read_settings_file(
                settings_filename,
                mandatory=True,
                include_stack=True,
                configs_dir_list=[model_dir(example_name, c) for c in config_dirs],
            )
        except Exception:
            logger.error(f"os.getcwd:{os.getcwd()}")
            raise
        if "models" not in existing_settings:
            raise ValueError(
                f"missing list of models from {config_dirs}/{settings_filename}"
            )
        models = existing_settings["models"]
        use_multiprocess = existing_settings.get("multiprocess", False)
        for k in existing_settings:
            print(f"existing_settings {k}:", existing_settings[k])

        settings_changes = dict(
            benchmarking=True,
            checkpoints=True,
            trace_hh_id=None,
            chunk_training_mode="disabled",
            inherit_settings=True,
        )
        settings_changes.update(benchmark_settings)

        # Pre-run checkpointing or Multiprocess timing runs only need to
        # include models up to the penultimate component to be benchmarked.
        last_component_to_benchmark = 0
        for cname in component_names:
            try:
                last_component_to_benchmark = max(
                    models.index(cname), last_component_to_benchmark
                )
            except ValueError:
                if cname not in models:
                    logger.warning(
                        f"want to benchmark {example_name}.{cname} but it is not in the list of models to run"
                    )
                else:
                    raise
        if use_multiprocess:
            last_component_to_benchmark += 1
        pre_run_model_list = models[:last_component_to_benchmark]
        if skip_component_names is not None:
            for cname in skip_component_names:
                if cname in pre_run_model_list:
                    pre_run_model_list.remove(cname)
        settings_changes["models"] = pre_run_model_list

        if "multiprocess_steps" in existing_settings:
            multiprocess_steps = existing_settings["multiprocess_steps"]
            while (
                multiprocess_steps[-1].get("begin", "missing-begin")
                not in pre_run_model_list
            ):
                multiprocess_steps = multiprocess_steps[:-1]
                if len(multiprocess_steps) == 0:
                    break
            settings_changes["multiprocess_steps"] = multiprocess_steps

        with open(
            model_dir(example_name, config_overload_dir, settings_filename), "wt"
        ) as yf:
            try:
                yaml.safe_dump(settings_changes, yf)
            except Exception:
                logger.error(f"settings_changes:{str(settings_changes)}")
                logger.exception("oops")
                raise
        with open(
            model_dir(example_name, config_overload_dir, "network_los.yaml"), "wt"
        ) as yf:
            benchmark_network_los["inherit_settings"] = True
            yaml.safe_dump(benchmark_network_los, yf)

        os.makedirs(model_dir(example_name, output_dir), exist_ok=True)

        # Running the model through all the steps and checkpointing everywhere is
        # expensive and only needs to be run once.  Once it is done we will write
        # out a completion token file to indicate to future benchmark attempts
        # that this does not need to be repeated.  Developers should manually
        # delete the token (or the whole model file) when a structural change
        # in the model happens such that re-checkpointing is needed (this should
        # happen rarely).
        use_config_dirs = (config_overload_dir, *config_dirs)
        token_file = model_dir(example_name, output_dir, "benchmark-setup-token.txt")
        if not os.path.exists(token_file) and not use_multiprocess:
            try:
                pre_run(
                    model_dir(example_name),
                    use_config_dirs,
                    data_dir,
                    output_dir,
                    settings_filename,
                )
            except Exception as err:
                with open(
                    model_dir(example_name, output_dir, "benchmark-setup-error.txt"),
                    "wt",
                ) as f:
                    f.write(f"error {err}")
                    f.write(traceback.format_exc())
                raise
            else:
                with open(token_file, "wt") as f:
                    # We write the commit into the token, in case that is useful
                    # to developers to decide if the checkpointed pipeline is
                    # out of date.
                    asv_commit = os.environ.get("ASV_COMMIT", "ASV_COMMIT_UNKNOWN")
                    f.write(asv_commit)
        if use_multiprocess:
            # Multiprocessing timing runs are actually fully completed within
            # the setup_cache step, and component-level timings are written out
            # to log files by activitysim during this run.
            asv_commit = os.environ.get("ASV_COMMIT", "ASV_COMMIT_UNKNOWN")
            try:
                pre_run(
                    model_dir(example_name),
                    use_config_dirs,
                    data_dir,
                    output_dir,
                    settings_filename,
                )
                run_multiprocess()
            except Exception as err:
                with open(
                    model_dir(
                        example_name, output_dir, f"-mp-run-error-{asv_commit}.txt"
                    ),
                    "wt",
                ) as f:
                    f.write(f"error {err}")
                    f.write(traceback.format_exc())
                raise

    except Exception as err:
        logger.error(
            f"error in template_setup_cache({example_name}):\n" + traceback.format_exc()
        )
        raise


def template_component_timings(
    module_globals,
    component_names,
    example_name,
    config_dirs,
    data_dir,
    output_dir,
    preload_injectables,
    repeat_=(1, 20, 10.0),  # min_repeat, max_repeat, max_time_seconds
    number_=1,
    timeout_=36000.0,  # ten hours,
    version_="1",
):
    """
    Inject ComponentTiming classes into a module namespace for benchmarking a model.

    Arguments with a trailing underscore get passed through to airspeed velocity, see
    https://asv.readthedocs.io/en/stable/benchmarks.html?highlight=repeat#timing-benchmarks
    for more info on these.

    Parameters
    ----------
    module_globals : Mapping
        The module globals namespace, into which the timing classes are written.
    component_names : Iterable[str]
        Names of components to benchmark.
    example_name : str
        Name of the example model being benchmarked, as it appears in the
        exammple_manifest.yaml file.
    config_dirs : Tuple[str]
        Config directories to use when running the model being benchmarked.
    data_dir, output_dir : str
        Data and output directories to use when running the model being
        benchmarked.
    preload_injectables : Tuple[str]
        Names of injectables to pre-load (typically skims).
    repeat_ : tuple
        The values for (min_repeat, max_repeat, max_time_seconds).  See ASV docs
        for more information.
    number_ : int, default 1
        The number of iterations in each sample.  Generally this should stay
        set to 1 for ActivitySim timing.
    timeout_ : number, default 36000.0,
        How many seconds before the benchmark is assumed to have crashed.  The
        typical default for airspeed velocity is 60 but that is wayyyyy too short for
        ActivitySim, so the default here is set to ten hours.
    version_ : str
        Used to determine when to invalidate old benchmark results. Benchmark results
        produced with a different value of the version than the current value will be
        ignored.
    """

    for componentname in component_names:

        class ComponentTiming:
            component_name = componentname
            warmup_time = 0
            min_run_count = 1
            processes = 1
            repeat = repeat_
            number = number_
            timeout = timeout_

            def setup(self):
                setup_component(
                    self.component_name,
                    model_dir(example_name),
                    preload_injectables,
                    config_dirs,
                    data_dir,
                    output_dir,
                )

            def teardown(self):
                teardown_component(self.component_name)

            def time_component(self):
                run_component(self.component_name)

            time_component.pretty_name = f"{example_name}:{componentname}"
            time_component.version = version_

        ComponentTiming.__name__ = f"{componentname}"

        module_globals[componentname] = ComponentTiming


def template_component_timings_mp(
    module_globals,
    component_names,
    example_name,
    output_dir,
    pretty_name,
    version_="1",
):
    """
    Inject ComponentTiming classes into a module namespace for benchmarking a model.

    This "MP" version for multiprocessing doesn't actually measure the time taken,
    but instead it parses the run logs from a single full run of the mode, to
    extract the per-component timings.  Most of the configurability has been removed
    compared to the single-process version of this function.

    Parameters
    ----------
    module_globals : Mapping
        The module globals namespace, into which the timing classes are written.
    component_names : Iterable[str]
        Names of components to benchmark.
    example_name : str
        Name of the example model being benchmarked, as it appears in the
        exammple_manifest.yaml file.
    output_dir : str
        Output directory to use when running the model being benchmarked.
    pretty_name : str
        A "pretty" name for this set of benchmarks.
    version_ : str
        Used to determine when to invalidate old benchmark results. Benchmark results
        produced with a different value of the version than the current value will be
        ignored.
    """

    for componentname in component_names:

        class ComponentTiming:
            component_name = componentname

            def track_component(self):
                durations = []
                inject.add_injectable("output_dir", model_dir(example_name, output_dir))
                logfiler = config.log_file_path(f"timing_log.mp_households_*.csv")
                for logfile in glob.glob(logfiler):
                    df = pd.read_csv(logfile)
                    dfq = df.query(f"component_name=='{self.component_name}'")
                    if len(dfq):
                        durations.append(dfq.iloc[-1].duration)
                if len(durations):
                    return np.mean(durations)
                else:
                    raise ValueError("no results available")

            track_component.pretty_name = f"{pretty_name}:{componentname}"
            track_component.version = version_
            track_component.unit = "s"

        ComponentTiming.__name__ = f"{componentname}"

        module_globals[componentname] = ComponentTiming



# This script looks for errors in the examples creates by `create_run_all_examples.py`

import argparse
import glob

parser = argparse.ArgumentParser()
parser.add_argument(
    "working_dir",
    type=str,
    metavar="PATH",
    help="path to examples working directory",
)

args = parser.parse_args()

files_with_errors = []

for logfile in glob.glob(f"{args.working_dir}/*/output/log/activitysim.log"):
    with open(logfile, "rt") as f:
        printing_traceback = False
        found_traceback = False
        for n, line in enumerate(f.readlines(), start=1):
            if printing_traceback:
                print(line.rstrip())
                if not line.startswith(" "):
                    printing_traceback = False
            else:
                if "Traceback" in line:
                    print(f"======= TRACEBACK in {logfile} at line {n} =======")
                    print(line.rstrip())
                    printing_traceback = True
                    found_traceback = True
        if not found_traceback:
            print(f"OK: {logfile}")
        else:
            files_with_errors.append(logfile)

if files_with_errors:
    print("=====================================================")
    print(f"Found {len(files_with_errors)} examples with errors:")
    for f in files_with_errors:
        print(f"- {f}")
    print("=====================================================")



# ActivitySim
# See full license in LICENSE.txt.
import os
import subprocess

import pandas as pd
import pandas.testing as pdt
import pkg_resources

from activitysim.core import inject


def teardown_function(func):
    inject.clear_cache()
    inject.reinject_decorated_tables()


def test_prototype_mtc_extended():
    def example_path(dirname):
        resource = os.path.join("examples", "prototype_mtc_extended", dirname)
        return pkg_resources.resource_filename("activitysim", resource)

    def example_mtc_path(dirname):
        resource = os.path.join("examples", "prototype_mtc", dirname)
        return pkg_resources.resource_filename("activitysim", resource)

    def test_path(dirname):
        return os.path.join(os.path.dirname(__file__), dirname)

    def regress():
        regress_trips_df = pd.read_csv(test_path("regress/final_trips.csv"))
        final_trips_df = pd.read_csv(test_path("output/final_trips.csv"))

        regress_vehicles_df = pd.read_csv(test_path("regress/final_vehicles.csv"))
        final_vehicles_df = pd.read_csv(test_path("output/final_vehicles.csv"))

        # person_id,household_id,tour_id,primary_purpose,trip_num,outbound,trip_count,purpose,
        # destination,origin,destination_logsum,depart,trip_mode,mode_choice_logsum
        # compare_cols = []
        pdt.assert_frame_equal(final_trips_df, regress_trips_df)
        pdt.assert_frame_equal(final_vehicles_df, regress_vehicles_df)

    file_path = os.path.join(os.path.dirname(__file__), "simulation.py")

    subprocess.run(
        [
            "coverage",
            "run",
            "-a",
            file_path,
            "-c",
            test_path("configs"),
            "-c",
            example_path("configs"),
            "-c",
            example_mtc_path("configs"),
            "-d",
            example_mtc_path("data"),
            "-o",
            test_path("output"),
        ],
        check=True,
    )

    regress()


if __name__ == "__main__":

    test_prototype_mtc_extended()
